from nibabel.freesurfer.io import read_annot, read_geometry
import numpy as np
import nibabel as nib
import pandas as pd
import os
from ..misc.text import get_unique_str_markers
from ..misc.print import _get_print
from .. import data_dr as def_data_dr
from ..loading.from_data import get_surf_loc
from difflib import SequenceMatcher
from ..misc.text import clean_key
from itertools import permutations

import warnings
with warnings.catch_warnings():
    warnings.simplefilter(action='ignore', category=FutureWarning)
    from nilearn.surface import load_surf_data

# Custom exception
class RepeatROIError(Exception):
    pass

def similar(a, b):
    return SequenceMatcher(None, a, b).ratio()

def _save_mapping(mapping, loc):
    
    with open(loc, 'w') as f:
        for m in mapping:
            f.write(str(m))
            f.write(',')
            f.write(str(mapping[m]))
            f.write('\n')

def _load_mapping(loc):

    mapping = {}

    with open(loc, 'r') as f:
        for line in f.readlines():
            line = line.split(',')
            mapping[line[0]] = line[1].rstrip()

    return mapping

def _get_col_names(df):
    
    all_cols = list(df)

    for a in all_cols:
        try:
            df[a].astype('float')
            value_col = a
        except ValueError:
            name_col = a

    return name_col, value_col

def _data_to_dict(data):

    # If series convert to dict directly
    if isinstance(data, pd.Series):
        return data.to_dict()
        
    # If df
    elif isinstance(data, pd.DataFrame):

        # If only 1 col, use index as name col
        if len(list(data)) == 1:
            return data[list(data)[0]].to_dict()

        # If 2 col DF proc
        as_dict = {}
        name_col, value_col = _get_col_names(data)

        for i in data.index:
            name = data[name_col].loc[i]
            as_dict[name] = data[value_col].loc[i]

        return as_dict

    elif isinstance(data, dict):
        return data

    raise RuntimeError('Invalid format passed for data, must be dict, pd.Series or pd.DataFrame')

def auto_determine_lh_rh_keys(data):

    # Use static key patterns to search for
    from ..loading.from_data import LH_KEYS, RH_KEYS
    
    # Get feat names as list from data
    feat_names = list(_data_to_dict(data).keys())

    # Compute tolerance
    tol = .15
    tol_n = len(feat_names) * tol
    tol_upper = len(feat_names) // 2 + tol_n
    tol_lower = len(feat_names) // 2 - tol_n

    # Search through each option
    for lh, rh in zip(LH_KEYS, RH_KEYS):

        # Check startswith, then endswith, then just is in
        for key_type in ['start', 'end', None]: 

            # Count number in each, depending on key type
            if key_type is None:
                n_lh = len([f for f in feat_names if lh in f])
                n_rh = len([f for f in feat_names if rh in f])
            elif key_type == 'start':
                n_lh = len([f for f in feat_names if f.startswith(lh)])
                n_rh = len([f for f in feat_names if f.startswith(rh)])
            elif key_type == 'end':
                n_lh = len([f for f in feat_names if f.endswith(lh)])
                n_rh = len([f for f in feat_names if f.endswith(rh)])
        
            # Should be around same size, but 
            # doesn't have to be exactly equal if
            # extending to other ROIs in future
            if n_lh > tol_lower and n_lh < tol_upper and n_rh > tol_lower and n_rh < tol_upper:
                return (lh, key_type), (rh, key_type)

    raise RuntimeError('Could not auto determine lh and rh unique keys.')

def _is_in(name, key):

    # Unpack tuple key
    k, k_type = key

    # Base is in
    if k_type is None:
        return k in name
    elif k_type == 'start':
        return name.startswith(k)
    elif k_type == 'end':
        return name.endswith(k)
    else:
        raise RuntimeError('Invalid key type.')

def _replace(name, key, new):

      # Unpack tuple key
    k, k_type = key

    if k_type is None:
        return name.replace(k, new)
    elif k_type == 'start':
        if name.startswith(k):
            return name[len(k):]
        return name
    elif k_type == 'end':
        if name.endswith(k):
            return name[:len(k)]
        return name
    else:
        raise RuntimeError('Invalid key type.')

def _get_pieces(key):
    return set([k for k in key.split(' ') if len(k) > 0])

def _get_roi_dict(data, i_keys=None, d_keys=None):
    '''If data is a df, with assume that the data is stored in two columns
    with name of ROI in one, and value in the other. Otherwise, if data is 
    a dictionary, assume that it is already in name of ROI, value form.

    i_keys is a list of keys where all of the entries if any have to present
    in the name of the roi for it to be kept, d_keys is a list / single item
    where if any of the d_keys are present in an roi it will be dropped.
    '''

    # Get as dict
    data = _data_to_dict(data)

    # Make a dict processing the passed i and d keys
    spec_data = {}
    for name in data:

        # In order to add, must qualify for all i_keys, and not any d_keys
        if all([_is_in(name, key) for key in i_keys]) and not any([_is_in(name, key) for key in d_keys]):
            spec_data[name] = data[name]

    # Generate list of alt names
    try:
        u_markers = get_unique_str_markers(list(spec_data))

    # If fails, then pass same as base
    except IndexError:
        u_markers = list(spec_data)

    return spec_data, u_markers

def _check_combos(not_found, name_pieces):
    
    # If size 0 or size 1, then return
    # as combo search is over
    if len(not_found) <= 1:
        return not_found
    
    # Generate all possible combinations, checking sequentially
    for n in range(2, len(not_found)+1):
        ps = permutations(not_found, n)
        
        # Check each combo
        for p in ps:
            combo = ''.join(p)
            
            # If present, call recursively
            # but with each piece removed from
            # not found
            if combo in name_pieces:
                for name in p:
                    not_found.remove(name)
                    
                return _check_combos(not_found, name_pieces)

    # If no matches, return as is
    return not_found

class Ref():
    
    def __init__(self, space, parc, data_dr='default', verbose=0, _print=None):
               
        self.space = space
        self.parc = parc

        # Get verbose print object
        self._print = _get_print(verbose, _print=_print)

        if data_dr == 'default':
            data_dr = def_data_dr

        self.data_dr = data_dr
        
        self._load_mappings()
        self._load_ref()
        
    def _load_mappings(self):

        if self.parc is None:
            self.mapping, self.label_2_int = None, None
            return

        map_loc = os.path.join(self.data_dr, 'mappings', self.parc + '.')

        try:
            self.mapping = _load_mapping(map_loc + 'mapping.txt')
            self.label_2_int = _load_mapping(map_loc + 'label_2_int.txt')
            self._print(f'Loaded mapping and label_2_int dicts, from: {map_loc}', level=2)
        except FileNotFoundError:
            self.mapping, self.label_2_int = None, None
    
    def _load_ref(self):
        pass
    
    def _get_ref_vals(self, hemi=None):
        pass

    def _add_to_inds(self, inds, name, label, trans_label):

        sim = similar(trans_label, name)
        ind = int(self.label_2_int[label])
        inds.append((ind, label, sim))
        self._print(f'Found match {trans_label} w/ similarity={sim}', level=3)

    def _proc_keys_input(self, keys):

        if keys is None:
            keys = []

        elif isinstance(keys, str):
            keys = [keys]

        elif isinstance(keys, tuple):
            if len(keys) != 2:
                raise RuntimeError(f'Invalid passed for keys={keys}')
            keys = [keys]

        elif isinstance(keys, list):
            for i, key in enumerate(keys):

                if isinstance(key, tuple) and len(key) == 2:
                    if key[1] not in [None, 'start', 'end']:
                        raise RuntimeError(f'Passed {key} if tuple, 2nd element must be None, start or end.')
                
                elif not isinstance(key, str):
                    raise RuntimeError(f'Passed element in keys {key} is not a valid str.')
                
                # If here, than the key is a single str, wrap in tuple w/ None
                else:
                    keys[i] = (key, None)
        
        else:
            raise RuntimeError(f'keys must be str or list of str / tuple.')

        return keys

    def _check_for_pieces_ind(self, name, check_combos=True):

        # Alternate case if sub-string not found
        # designed to get at cases where the order might be wrong
        inds = []
        
        # Get names as pieces representation
        name_pieces = _get_pieces(name)

        for label in self.label_2_int:

            # Use the cleaned version of each
            trans_label = clean_key(label)
            trans_pieces = _get_pieces(trans_label)

            # Compute intersection
            piece_intersection = trans_pieces.intersection(name_pieces)

            # If the same
            if len(piece_intersection) == len(trans_pieces):
                self._add_to_inds(inds, name, label, trans_label)

            # Otherwise, check recursively for combinations
            elif check_combos:

                not_found = trans_pieces - piece_intersection
                not_found = _check_combos(not_found, name_pieces)
                
                # If empty, found
                if len(not_found) == 0:
                    self._add_to_inds(inds, name, label, trans_label)

        return inds

    def _check_for_ind(self, name, original_name):

        # Try to find the correct index
        inds = []

        # These labels are the ground truth we need
        # to map to the saved index
        for label in self.label_2_int:

            # Use the cleaned version of each
            trans_label = clean_key(label)

            # If the ground truth is sub-string of the current name
            # Keep track, w/ also a simmilarity score
            if trans_label in name:
                ind = int(self.label_2_int[label])

                # Append ind, label and simmilarity
                self._add_to_inds(inds, name, label, trans_label)

        # If none found from base index search, use pieces method
        if len(inds) == 0:
            inds  = self._check_for_pieces_ind(name)

        # If more than one, sort so first is highest simmilarity
        if len(inds) > 1:
            inds = sorted(inds, key=lambda i : i[2], reverse=True)

        #  Return first, if only one then that is first
        if len(inds) > 0:
            self._print(f'Mapping: {original_name} -> {inds[0][1]}.', level=2)
            return inds[0][0]

        # If not found, return None
        return None

    def _apply_mappings_to_name(self, name):

        self._print(f'Applying mapping to {name}', level=2)

        # Apply the mapping for this parcellation
        for key in self.mapping:

            # Get a clean representation of a possible transform key
            trans_key = clean_key(key)

            # If found in the current ROI name, replace it
            if trans_key in name:

                # Get the cleaned new value to replace with
                new = clean_key(self.mapping[key])
                name = name.replace(trans_key, new)
                self._print(f'replace: "{trans_key}":"{new}", new="{name}"', level=3)

        return name
        

    def _find_ref_ind(self, name, alt_name=None, i_keys=None):

        # Keep copy of original name for verbose print
        original_name = name
        
        # Base clean name + store copy
        name = clean_key(name)
        clean_name = name
        self._print(f'clean_key: "{original_name}" to "{name}"', level=3)

        # Try applying mappings to name, then perform first check 
        name = self._apply_mappings_to_name(name)
        ind = self._check_for_ind(name, original_name)
        if ind is not None:
            return ind

        # Next, let's try with the pre-mapped name
        ind = self._check_for_ind(clean_name, original_name)
        if ind is not None:
            return ind

        # Next, try again but with all i_keys removed from the name
        if i_keys is not None:
            next_name = original_name
            for key in i_keys:
                next_name = _replace(next_name, key, '')

            self._print(f'Did not find roi for "{original_name}", trying with "{next_name}"', level=2)
            
            # We make sure to pass i_keys as None here, to avoid infinite loops
            ind = self._find_ref_ind(next_name, alt_name=alt_name, i_keys=None)
            if ind is not None:
                return ind

        # If still haven't found anything, try everything again but with alternate name
        if alt_name is not None:
            self._print(f'Did not find roi for "{name}", trying with "{alt_name}"', level=2)

            # We make sure to pass alt_name as None here, to avoid infinite loops
            ind = self._find_ref_ind(alt_name, alt_name=None, i_keys=i_keys)
            if ind is not None:
                return ind

        # If finally not found, return None
        return None

    def find_ref_ind(self, name, alt_name=None, i_keys=None):

        # Try to find ind
        ind = self._find_ref_ind(name=name, alt_name=alt_name, i_keys=i_keys)
        if ind is not None:
            return ind

         # If error, print out the true region names
        self._print('The passed values must be able to map in some way to one of these reference region names:',
                    list(self.label_2_int), 'You can turn on verbose to see what is already being checked.', level=0)

        # If didn't find with alt name also
        raise RuntimeError(f'Could not find matching ROI for: "{name}"')
    
    def _get_plot_vals(self, roi_dict,
                       roi_alt_names,
                       ref_vals, i_keys, use_just_alt=False):

         # Init plot vals
        plot_vals = np.zeros(np.shape(ref_vals))

        # Keep track of already found inds
        used_inds = set()
        for name, alt_name in zip(roi_dict, roi_alt_names):

            # Get plot value
            value = roi_dict[name]

            # Optionally, override name w/ alt_name
            if use_just_alt:
                name = alt_name

            # Try to find the name in the reference values
            ind = self.find_ref_ind(name=name, alt_name=alt_name, i_keys=i_keys)

            # Check to see if this ROI is mapping to 
            # one another has already mapped to,
            # if so, trigger an error.
            if ind in used_inds:
                self._print('Overlapping ROI mapping error.', level=2)
                raise RepeatROIError(f'Error: mapping {name} failed, as another ROI already mapped to the same reference (ind={ind}). '\
                                      'Set verbose>=2 in Ref object and double check how each ROI is being mapped to make sure it is correct. ')
            else:
                used_inds.add(ind)

            # Set the proper values based on the found index
            plot_vals = np.where(ref_vals == ind, value, plot_vals)

        return plot_vals

    def get_plot_vals(self, data, hemi=None, i_keys=None, d_keys=None):

        self._print(f'Start get plot vals for hemi={hemi}', level=1)

        # Process keys input
        i_keys, d_keys = self._proc_keys_input(i_keys), self._proc_keys_input(d_keys)
        
        # Get base roi dict
        roi_dict, roi_alt_names = _get_roi_dict(data, i_keys, d_keys)
        self._print(f'roi_alt_names={roi_alt_names}', level=3)
        
        # Load reference values
        ref_vals = self._get_ref_vals(hemi)

        # Try - but in case of repeat ROI error
        # then try again, but with use_just_alt
        try:
            plot_vals = self._get_plot_vals(roi_dict, roi_alt_names,
                                            ref_vals, i_keys, use_just_alt=False)
        except RepeatROIError:
            self._print(f'Trying plot vals w/ use_just_alt=True', level=2)
            plot_vals = self._get_plot_vals(roi_dict, roi_alt_names,
                                            ref_vals, i_keys, use_just_alt=True)
       
        return plot_vals

class SurfRef(Ref):
    
    def __init__(self, space='fsaverage5', parc='destr',
                 data_dr='default', surf_mesh=None,
                 bg_map=None, verbose=0, _print=None):

        if space == 'default':
            space = 'fsaverage5'

        super().__init__(space, parc, data_dr, verbose=verbose, _print=_print)

        self.surf_mesh = surf_mesh
        self.bg_map = bg_map
    
    def _load_ref(self):
        
        # If no parc, skip
        if self.parc is None:
            return

        ref_loc = os.path.join(self.data_dr, self.space, 'label')
        self._print(f'Using base ref_loc = {ref_loc}', level=2)

        # Try to load lh
        lh_loc = os.path.join(ref_loc, 'lh.' + self.parc)

        if os.path.exists(lh_loc + '.annot'):
            self.lh_ref = read_annot(lh_loc + '.annot')[0]
        elif os.path.exists(lh_loc + '.gii'):
            self.lh_ref = load_surf_data(lh_loc + '.gii')
        elif os.path.exists(lh_loc + '.npy'):
            self.lh_ref = np.load(lh_loc + '.npy')
        else:
            raise RuntimeError('Could not find lh ref parcel for this space')

        # Try to load rh
        rh_loc = os.path.join(ref_loc, 'rh.' + self.parc)
        if os.path.exists(rh_loc + '.annot'):
            self.rh_ref = read_annot(rh_loc + '.annot')[0]
        elif os.path.exists(rh_loc + '.gii'):
            self.rh_ref = load_surf_data(rh_loc + '.gii')
        elif os.path.exists(rh_loc + '.npy'):
            self.rh_ref = np.load(rh_loc + '.npy')
        else:
            raise RuntimeError('Could not find rh ref parcel for this space')

    def _get_ref_vals(self, hemi):
        
        if hemi == 'lh' or hemi == 'left':
            ref_vals = self.lh_ref
        else:
            ref_vals = self.rh_ref
    
        return ref_vals
    
    def get_hemis_plot_vals(self, data, lh_key='auto', rh_key='auto', i_keys=None, d_keys=None):

        # Check  to see if either lh_key or rh_key is passed as auto
        if lh_key == 'auto' or rh_key == 'auto':
            
            # Get auto determined
            lh_a_key, rh_a_key = auto_determine_lh_rh_keys(data)
            self._print(f'Auto determined keys as lh_key={lh_a_key}, rh_key={rh_a_key}.', level=1)

            # Only override passed value for one if auto
            if lh_key == 'auto':
                lh_key = lh_a_key
            if rh_key == 'auto':
                rh_key = rh_a_key

        # Process input
        i_keys, d_keys = self._proc_keys_input(i_keys), self._proc_keys_input(d_keys)
        
        # Get the plot values per hemisphere seperately
        lh_plot_vals = self.get_plot_vals(data, 'lh', i_keys+[lh_key], d_keys)
        rh_plot_vals = self.get_plot_vals(data, 'rh', i_keys+[rh_key], d_keys)
        
        return {'lh': lh_plot_vals, 'rh': rh_plot_vals}

    def get_surf(self, name, hemi):

        if name is None:
            return None
        
        # If already surf mesh like
        if not isinstance(name, str):
            return name
        
        # Assume name is path if exists
        if os.path.exists(name):
            loc = name

        # Otherwise find right loc if any
        else:
            loc = get_surf_loc(space=self.space, hemi=hemi, key=name)
            if loc is None:
                raise  RuntimeError(f'Unable to find space={self.space}, hemi={hemi}, name={name}')

        # If gifti
        if loc.endswith('.gii'):
            surf = load_surf_data(loc)
            if len(surf) == 2:
                surf = (surf[0], surf[1])
            return surf

        # Otherwise try geometry first then base surf data
        try:
            return read_geometry(loc)
        except ValueError:
            return load_surf_data(loc)

class VolRef(Ref):
    
    def __init__(self, space='mni_1mm', parc='aseg', data_dr='default',
                 verbose=0, _print=None):

        if space == 'default':
            space = 'mni_1mm'

        super().__init__(space, parc, data_dr, verbose=verbose, _print=_print)
    
    @property
    def shape(self):
        return self.ref_vol.shape
    
    def _load_ref(self):
        
        ref_loc = os.path.join(self.data_dr, self.space)
        self._print(f'Using base ref_loc = {ref_loc}', level=2)
        
        # Get w/ flexible to different file extension
        options = os.listdir(ref_loc)
        options_no_ext = [o.split('.')[0] for o in options]

        try:
            ind = options_no_ext.index(self.parc)

        # If doesn't exist, help user
        except ValueError:
            self._print(f'Note valid parc options for {self.space} are:', options_no_ext, level=0)
            raise RuntimeError(f'Space: {self.space} does not have parc {self.parc}.')

        # Load with nibabel, since need affine
        ref_vol_raw = nib.load(os.path.join(ref_loc, options[ind]))

        # Save to class
        self.ref_vol_affine = ref_vol_raw.affine
        self.ref_vol = np.array(ref_vol_raw.get_fdata())
        
    def _get_ref_vals(self, hemi=None):
        return self.ref_vol
    
    def get_plot_vals(self, data, hemi=None, i_keys=[], d_keys=[]):
        
        plot_vals = super().get_plot_vals(data, hemi, i_keys, d_keys)
        return nib.Nifti1Image(plot_vals, self.ref_vol_affine)


def _get_vol_ref_from_guess(voxel_inds):

    # Just these options for now - any others should add? TODO
    # Note: Favors hcp_rois parc since those are cifti native

    # MNI 2mm
    if np.max(voxel_inds) == 76:
        return VolRef(space='mni_2mm', parc='hcp_rois')
   
    # MNI 1.6mm
    elif np.max(voxel_inds) == 95:
        return VolRef(space='mni_1.6mm', parc='hcp_rois')

    # MNI 1mm as base case
    return VolRef(space='mni_1mm', parc='aseg')

    
    









from pathlib import Path
import os
from urllib import request
import tarfile
import warnings

# Set version
__version__ = '0.22'
__data_version__ = '1.2.5'

CURRENT_DR = Path().resolve()
DATA_REF_LOC = os.path.join(CURRENT_DR, 'data_ref.txt')

def get_data_dr_path(download_loc):
    return os.path.join(download_loc, f'neurotools_data-{__data_version__}', 'data')

def download(download_loc):

    print(f'Downloading latest neurotools_data to {download_loc}')
    zipped_file = f'{download_loc}.tar.gz'

    # Download
    request.urlretrieve(f'https://github.com/sahahn/neurotools_data/archive/refs/tags/{__data_version__}.tar.gz', zipped_file)

    # Make sure main directory exists
    os.makedirs(download_loc, exist_ok=True)

    # Un-zip to main directory
    with tarfile.open(zipped_file) as f:
        f.extractall(download_loc)

    # Remove tar gz file once done
    os.remove(zipped_file)

    # Save location of the main neurotools data file 
    with open(DATA_REF_LOC, 'w') as f:
        f.write(download_loc)

    # This is the location of the download data
    data_dr = get_data_dr_path(download_loc)

    print(f'Downloaded data version = {__data_version__} complete!')
    print(f'Current version saved at: {data_dr}')
    print(f'If you move this directory, make sure to update saved location in data ref at {DATA_REF_LOC}.')

    return data_dr

def check_DATA_REF_LOC():

    # If doesn't exist, init
    if not os.path.exists(DATA_REF_LOC):
        
        # Set to default download loc if doesn't already exist
        # TODO could prompt user instead?
        home_dr = str(Path.home())
        download_loc = os.path.join(home_dr, 'neurotools_data')

        # Save default
        with open(DATA_REF_LOC, 'w') as f:
            f.write(download_loc)

def resolve_data_dr():

    # Make sure data ref has been init'ed
    check_DATA_REF_LOC()

    # Check saved    
    with open(DATA_REF_LOC, 'r') as f:
        download_loc = f.readline().rstrip()

    # If exists, check to see if latest version exists
    data_dr = get_data_dr_path(download_loc)
    if os.path.exists(data_dr):
        return data_dr

    # Otherwise, download latest
    return download(download_loc)

# Resolve data dr, downloading if needed
data_dr = resolve_data_dr()





import numpy as np
import pytest
from ..rois import SurfLabels, SurfMaps

def test_surf_labels():

    sl = SurfLabels(labels=np.array([1, 1, 2, 2]))
    data = np.array([5, 5, 10, 10])

    Xt = sl.fit_transform(data)
    assert np.array_equal(Xt, np.array([5, 10]))

    reverse_trans = sl.inverse_transform(Xt)
    assert np.array_equal(reverse_trans, np.array([5, 5, 10, 10]))

def test_surf_labels2():

    sl = SurfLabels(labels=np.array([0, 2, 2, 2, 3, 3]))
    data = np.array([1, 2, 3, 4, 5, 6])

    Xt = sl.fit_transform(data)
    assert np.array_equal(Xt, np.array([3, 5.5]))

    reverse_trans = sl.inverse_transform(Xt)
    assert np.array_equal(reverse_trans, np.array([0, 3, 3, 3, 5.5, 5.5]))

def test_surf_labels_mask():

    labels = np.array([1, 1, 2, 2])
    data = np.array([1, 2, 3, 4])

    mask = np.array([1, 0, 0, 0])
    Xt = SurfLabels(labels=labels, mask=mask,
                    strategy='mean').fit_transform(data)

    assert np.array_equal(Xt, np.array([2, 3.5]))


def test_bad_surf_labels_mask():

    labels = np.array([1, 1, 2, 2])
    data = np.array([1, 2, 3, 4])
    mask = np.array([1, 0, 0, 0, 0])

    with pytest.raises(RuntimeError):
        SurfLabels(labels=labels, mask=mask,
                   strategy='mean').fit_transform(data)


def test_surf_labels_dif_strats():

    labels = np.array([1, 1, 2, 2])
    data = np.array([1, 2, 3, 4])

    Xt = SurfLabels(labels=labels,
                    strategy='median').fit_transform(data)
    assert np.array_equal(Xt, np.array([1.5, 3.5]))

    Xt = SurfLabels(labels=labels,
                    strategy='sum').fit_transform(data)
    assert np.array_equal(Xt, np.array([3, 7]))

    Xt = SurfLabels(labels=labels,
                    strategy='min').fit_transform(data)
    assert np.array_equal(Xt, np.array([1, 3]))

    Xt = SurfLabels(labels=labels,
                    strategy='max').fit_transform(data)
    assert np.array_equal(Xt, np.array([2, 4]))

    Xt = SurfLabels(labels=labels,
                    strategy='std').fit_transform(data)
    assert np.array_equal(Xt, np.array([.5, .5]))

    Xt = SurfLabels(labels=labels,
                    strategy='var').fit_transform(data)
    assert np.array_equal(Xt, np.array([.25, .25]))


def test_surf_data_2d():

    labels = np.array([1, 2, 1, 2])
    data = np.array([[1, 2, 3], [2, 3, 4],
                     [1, 2, 3], [2, 3, 4]])

    sl = SurfLabels(labels=labels)
    Xt = sl.fit_transform(data)
    assert np.array_equal(Xt, np.array([1, 2, 3, 2, 3, 4]))

    reverse_trans = sl.inverse_transform(Xt)
    assert np.array_equal(reverse_trans, np.array([[1, 2, 3],
                                                   [2, 3, 4],
                                                   [1, 2, 3],
                                                   [2, 3, 4]]))


def test_surf_data_2d_no_vectorize():

    labels = np.array([1, 2, 1, 2])
    data = np.array([[1, 2, 3], [2, 3, 4],
                     [1, 2, 3], [2, 3, 4]])

    sl = SurfLabels(labels=labels, vectorize=False)
    Xt = sl.fit_transform(data)
    assert np.array_equal(Xt, np.array([[1, 2, 3], [2, 3, 4]]))

    reverse_trans = sl.inverse_transform(Xt)
    assert np.array_equal(reverse_trans, np.array([[1, 2, 3],
                                                   [2, 3, 4],
                                                   [1, 2, 3],
                                                   [2, 3, 4]]))


def test_surf_labels_2d_fail():

    labels = np.array([[1, 1, 1], [2, 2, 2], [1, 1, 1], [2, 2, 2]])
    data = np.array([[1, 2, 3], [2, 3, 4], [1, 2, 3], [2, 3, 4]])

    sl = SurfLabels(labels=labels, vectorize=False)

    with pytest.raises(RuntimeError):
        sl.fit_transform(data)

def test_surf_maps_ls():

    maps = np.array([[0, 1],
                     [0, 2],
                     [1, 0],
                     [1, 0]])

    data = np.array([1.0, 1.0, 5.0, 5.0])
    data_dtype = data.dtype.name

    sm = SurfMaps(maps=maps, strategy='ls')
    Xt = sm.fit_transform(data)

    assert data_dtype == Xt.dtype.name
    np.testing.assert_almost_equal(Xt, np.array([5, .6]), decimal=6)

    # Reverse transform
    r_trans = sm.inverse_transform(Xt)
    np.array_equal(r_trans, np.array([.6, 1.2, 5, 5]))


def test_surf_maps_average():

    maps = np.array([[1, 0],
                     [1, 0],
                     [0, 1],
                     [0, 2]])

    data = np.array([1.0, 3.0, 1.0, 2.0])
    data_dtype = data.dtype.name

    sm = SurfMaps(maps=maps, strategy='average')
    Xt = sm.fit_transform(data)

    assert data_dtype == Xt.dtype.name
    np.array_equal(Xt, np.array([2, 1.66666667]))


def test_surf_maps_average_fail():

    maps = np.array([[1, 0],
                     [1, 0],
                     [0, 1],
                     [0, 2]])

    data = np.array([1.0, 3.0, 1.0, 2.0])

    sm = SurfMaps(maps=maps, strategy='average')
    Xt = sm.fit_transform(data)

    with pytest.raises(RuntimeError):
        sm.inverse_transform(Xt)


import nibabel as nib
import numpy as np
import os
import time
from joblib import Parallel, delayed
from nibabel import freesurfer as fs
from nibabel import GiftiImage
from .cache import _base_cache_load
from ..misc.print import _get_print

def _gifti_to_np(data):
    return np.asarray([arr.data for arr in data.darrays]).T.squeeze()

def load(f, index_slice=None, dtype=None):
    '''Smart loading function for neuro-imaging type data.

    Parameters
    ----------
    f : str, :func:`arrays<numpy.array>` or other
        The location / file path of the data to load passed
        as a str, or can accept directly numpy :func:`arrays<numpy.array>`
        or objects from nibabel, ect...

    index_slice : slices, tuple of slices, or :data:`None`, optional
        You may optional pass index slicing here.
        The typical benefit of index_slicing over masking in a later
        step is where data to load will be read with nibabel
        (e.g., saved as .nii or .mgz) and can therefore use an index_slice
        to load in only the data of interest. 

        For example let's say the shape of data is 5 x 20 on
        the disk, but you only need the data in the first 20, e.g.,
        so with traditional slicing that is [0], in this case by using the
        mask you have to load the full data from the disk, then slice it.
        But if you pass :python:`slice(0)` here, then this accomplishes the same thing,
        but only loads the requested slice from the disk.
        
        You must use python keyword `slice` for complex slicing. For example
        in the example above you may pass either :python:`(0)` or :python:`slice(0)`,
        but if you wanted something more complex, e.g. passing
        something like :python:`my_array[1:5:2, ::3]` you should pass
        :python:`(slice(1,5,2), slice(None,None,3))` here.

        ::

            default = None

    dtype : str or None, optional
        The data type in which to cast loaded data to.
        If left as None, keep original data type, otherwise
        can pass as a string name of datatype to cast to.
        For example, 'float32' to cast to 32 bit floating point
        precision.

        ::

            default = None

    Returns
    ---------
    data : :class:`array<numpy.ndarray>` or :data:`None`
        Returned is the requested data as a :func:`numpy array<numpy.array>`,
        or if passed :data:`None`, then :data:`None` will be returned. The shape
        of the returned data will be the shape of the data array portion of
        whatever data is passed with extra singular dimensions removed
        (i.e., np.squeeze is applied). If passed an index slice, only
        the portion of the data specified by the index slice will be loaded
        and returned.

        Note: The data type of the returned data will depend on how the
        data to be loaded is saved.

    Examples
    ---------
    Data can be loaded by simply passing a file location
    to load. For example, we can load just an :func:`array<numpy.array>` of ones:

    .. ipython:: python

        from neurotools.loading import load
        data = load('data/ones.nii.gz')
        data.shape

    This function is likewise robust to being passed non-file paths, for
    example passing an already defined :func:`array<numpy.array>` to load:

    .. ipython:: python

        import numpy as np

        data = np.ones((5, 3, 1))
        load(data).shape

    Note here that load will also apply a squeeze function to any loaded data,
    getting rid of any used dimensions.

    Next, let's look at an example where we pass a specific index slice.

    .. ipython:: python

        data = np.random.random((5, 3))
        
        # Let's say we want to select columns 0 and 1
        data[:, 0:2]
        
        # It's a bit more awkward but we pass:
        load(data, index_slice=(slice(None), slice(0, 2)))

    '''

    def _proc(data):
        '''Helper func, squeezes data and returns
        with correct data type'''

        if dtype is not None:
            return np.squeeze(data.astype(dtype))
        
        return np.squeeze(data)
    
    # If already numpy array, return of as is
    if type(f) is np.ndarray:
        raw = f.copy()
    
    # If gifti already
    elif isinstance(f, GiftiImage):
        raw = _gifti_to_np(f)
    
    # If nifti like
    elif isinstance(f, (nib.Cifti2Image, nib.Nifti1Image)):
        raw = f.get_fdata()

    # Keep as None if None
    elif f is None:
        return None

    # TODO add case for loading .mat files
    # from scipy.io import loadmat
    # ex: lh_d = loadmat(lh_loc)['parcels'].squeeze()
    
    # If loading from file
    elif isinstance(f, str):
    
        # Freesurfer cases (from nilearn load func)
        if (f.endswith('area')
            or f.endswith('curv')
            or f.endswith('sulc')
            or f.endswith('thickness')):
            raw = fs.io.read_morph_data(f)

        elif f.endswith('annot'):
            raw = fs.io.read_annot(f)[0]

        elif f.endswith('label'):
            raw = fs.io.read_label(f)
        
        # Numpy case
        elif f.endswith('.npy') or f.endswith('.npz'):
            raw = np.load(f)
        
        # Otherwise, try to load with nibabel
        else:

            data = nib.load(f)
            
            # If Gifti, handle special
            if isinstance(data, GiftiImage):
                raw = _gifti_to_np(data)
            
            # Special case if load with nibabel and index_slice is
            # passed, load from the dataobj
            elif index_slice is not None:
                
                # Return directly here
                return _proc(data.dataobj[index_slice])
            
            # Otherwise just load full
            else:
                raw = data.get_fdata()

    # If a Parc object
    elif hasattr(f, 'get_parc'):
        raw = f.get_parc(copy=True)

    else:
        raise RuntimeError(f'Unable to load passed {f}')

    # Return either array or index'ed array
    if index_slice is not None:
        return _proc(raw[index_slice])

    return _proc(raw)

def _apply_template(subject, template_path, contrast=None):

    # Base case is template path as a str
    if isinstance(template_path, str):
    
        # Copy template path under f
        f = template_path 

        # Replace subject
        f = f.replace('SUBJECT', str(subject))
        
        # Optionally replace contrast
        if contrast is not None:
            f = f.replace('CONTRAST', str(contrast))
        
        return f

    # Alternative case is that template path is passed as function
    if contrast is None:
        return template_path(subject)
    
    # Subject + contrast, func case
    return template_path(subject, contrast)

def _load_subject(subject, contrast, template_path, mask=None,
                  index_slice=None, dtype=None, _print=print):

    # Get the specific path for this subject based on the passed template
    f = _apply_template(subject=subject,
                        template_path=template_path,
                        contrast=contrast)

    _print('Loading:', f, level=2)
    raw = load(f, index_slice=index_slice, dtype=dtype)

    # If no mask, just flatten / make sure 1D
    if mask is None:
        flat_data = raw.flatten()
    
    # If mask, then use to create 1D array
    else:
        flat_data = raw[mask == 1]

    return flat_data
 
def _load_mask(mask):

    # If mask is None, return None, None
    if mask is None:
        return None, None

    # Check to see if mask already has affine
    # e.g., is already a NiftiImage
    if hasattr(mask, 'affine'):
        return _load_check_mask(mask, getattr(mask, 'affine'))
    
    # If the mask is a str - try to load it
    # with nibabel first, to check for affine
    if isinstance(mask, str):
        try:
            raw_mask = nib.load(mask)
            
            # If loaded with nibabel and has affine, return
            # data and affine
            if hasattr(raw_mask, 'affine'):
                return _load_check_mask(raw_mask, getattr(raw_mask, 'affine'))
        
        # All other cases, skip
        except:
            pass

    # If reaches here, then no affine
    # in this case, just load, then return as data with None affine
    return _load_check_mask(mask, None)

def _load_check_mask(mask, affine):

    mask_data = load(mask)
    n_unique = len(np.unique(mask_data))

    if n_unique != 2:
        raise RuntimeError(f'Passed mask must have only 2 unique values! Passed mask with {n_unique} unique values.')

    return mask_data.astype('bool'), affine

def _base_load_data(subjects, contrast, template_path, mask,
                    index_slice, dtype, n_jobs, verbose):

    # Init print based on verbose
    _print = _get_print(verbose)

    # Single core case
    if n_jobs == 1:

        subjs_data = []
        for subject in subjects:
            subjs_data.append(_load_subject(subject, contrast, template_path,
                                            mask=mask, index_slice=index_slice,
                                            dtype=dtype,
                                            _print=_print))

    # Multi-proc case
    else:
        subjs_data = Parallel(n_jobs=n_jobs, prefer="threads")(
            delayed(_load_subject)(
                subject=subject,
                contrast=contrast,
                template_path=template_path,
                mask=mask,
                index_slice=index_slice,
                dtype=dtype,
                _print=_print) for subject in subjects)

    # Return array with shape as the number of subjects by the sum of the mask
    data = np.stack(subjs_data)

    return data

def _cache_load_data(cache_loc, load_args):

    # TODO
    # Needs to be responsible here for any potential
    # re-ordering of subjects or anything - hmm not sure how actually
    # maybe make sure subjects isn't list-like, i.e., cast to array, so
    # that it doesn't cache based on that
    # subjects = load_args[0]

    data = np.load(cache_loc)
    return data

def _cache_save_data(data, cache_loc):
    np.save(cache_loc, data)

def load_data(subjects, template_path, contrast=None, mask=None,
              index_slice=None, zero_as_nan=False, nan_as_zero=False,
              dtype=None, n_jobs=1, verbose=1, _print=None, **cache_args):
    '''This method is designed to load data saved in a particular way,
    specifically where each subject / participants data is saved seperately.

    Parameters
    ----------
    subjects : array-like
        A list or array-like with the names of the subjects to load, where
        the names correspond in some way to the way the way the subject's data
        is saved. This correspondence is specified in template_path.

    template_path : str or func
        A str indicating the template form for how a single
        subjects data should be loaded, where SUBJECT will be
        replaced with that subjects name, and optionally CONTRAST will
        be replaced with the contrast name. 

        For example, to load subject X's contrast Y saved under
        'some_loc/X_Y.nii.gz'
        the template_path would be: 'some_loc/SUBJECT_CONTRAST.nii.gz'.

        You may also alternatively pass template_path as a python function,
        with first argument accepting subject, and optional second argument
        accepting contrast. In this case, the function should return the correct
        path for that subject / contrast pair when given one or both of these arguments.

        Note: The use of CONTRAST within the template path is optional, but
        this parameter is not.

    contrast : str or :data:`None`, optional
        The name of the contrast, used along with the template
        path to define where to load data.
        If passed :data:`None`, then it is assumed that CONTRAST
        is not present in the template_path and will be ignored.

        ::

            default = None

    mask : str, :func:`array<numpy.array>` or :data:`None`, optional
        After data is loaded, it can optionally be
        masked according to a specific :func:`array<numpy.array>`.

        If :data:`None`, and the data to be loaded is multi-dimensional,
        then the data will be flattened by default, e.g., if each data 
        point originally has shape 10 x 2, the flattened shape will be 20.
       
        If passed a str, it will be assumed to be the location of a mask
        in which to load, where the shape of the mask should match the data.

        When passing a mask, either by location or by :func:`array<numpy.array>` directly,
        values set to either 1 or True indicate the value should be kept,
        whereas values of 0 or False indicate the values at that location should
        be discarded.

        If a mask is used, see function :func:`funcs.reverse_mask_data`
        for reversing masked data (e.g., an eventual statistical output)
        back into it's pre-masked state.

        ::

            default = None

    index_slice : slices, tuple of slices, or :data:`None`, optional
        You may optional pass index slicing here.
        The typical benefit of index_slicing over masking
        is where data to load will be read with nibabel
        (e.g., saved as .nii or .mgz) and can therefore use an index_slice
        to load in only the data of interest. 

        For example let's say the shape of data is 5 x 20 on
        the disk, but you only need the data in the first 20, e.g.,
        so with traditional slicing that is [0], in this case by using the
        mask you have to load the full data from the disk, then slice it.
        But if you pass :python:`slice(0)` here, then this
        accomplishes the same thing,
        but only loads the requested slice from the disk.

        You must use python keyword `slice` for complex slicing. For example
        in the example above you may pass either :python:`(0)` or :python:`slice(0)`,
        but if you wanted something more complex, e.g. passing
        something like :python:`my_array[1:5:2, ::3]` you should pass
        :python:`(slice(1,5,2), slice(None,None,3))` here.

        If passed :data:`None` (the default), this option is
        ignored and the full data avaliable is loaded.

        ::

            default = None

    zero_as_nan : bool, optional
        Often in neuroimaging data, NaN values
        are encoded as 0's (I'm not sure why either).
        If this flag is set to True, then any 0's
        found in the loaded data will be replaced with NaN.
        
        ::

            default = False

    nan_as_zero : bool, optional
        As an alternative to zero_as_nan, can
        instead set any NaNs found to zeros.
        Note, if this is True then zero_as_nan can
        not also be True.

        ::

            default = False

    dtype : str or None, optional
        The data type in which to cast loaded data to.
        If left as None, keep original data type, otherwise
        can pass as a string name of datatype to cast to.
        For example, 'float32' to cast to 32 bit floating point
        precision.

        ::

            default = None

    n_jobs : int, optional
        The number of threads to use when loading data.

        Note: Often n_jobs will refer to number of separate processors,
        but in this case it refers to threads,
        which are not limited by the number of cores avaliable.

        ::

            default = 1

    verbose : int, optional
        This parameter controls the verbosity of this function.

        - If -1, then no message at all will be printed.
        - If 0, only warnings will be printed.
        - If 1, general status updates will be printed.
        - If >= 2, full verbosity will be enabled.

        ::

            default = 1

    cache_args : keyword arguments
        There are a number of optional cache arguments that can be set via
        kwargs, as listed below.

        - **cache_dr** : str or None
        
          The location of where to cache the results of
          this function, for faster loading in the future.

          If None, do not cache. The default if not set is
          'default', which will use caching in a location defined
          by the function name in a folder called
          neurotools_cache in the users homes directory.

        - **cache_max_sz** : str or int
        
          This parameter defines the maximize size of
          the cache directory. The idea is that if saving a new
          cached function call and it exceeds this cache max size,
          previous saved caches (by oldest in terms of used) will be
          deleted, ensuring the cache directory remains under this size.

          Can either pass in terms of bytes directly as a number,
          or in terms of a str w/ byte marker, e.g., '14G' for
          14 gigabytes, or '10 KB' for 10 kilobytes.

          The default if not set is '30G'.

        - **use_base_name** : bool
            
          Optionally when any arguments used in the
          caching can be cached based on either their full file
          path, if use_base_name is False, or just the file name itself,
          so for example /some/path/location vs. just location.
          The default if not set is True, as it assumes that maybe
          another file in another location with the same name is the same.

    Returns
    --------
    data : :class:`array<numpy.ndarray>`
        Loaded data across all specified subjects is returned
        as a 2D numpy :func:`array<numpy.array>` (shape= subj x data) where the first
        dimension is subject and the second is a single dimension representation
        of that subjects data.

    Examples
    ----------
    Consider the following simplified example where we are interested in loading and
    concatenating the data from three subjects: 'subj1', 'subj2' and subj3'.

    .. ipython:: python
        
        import os
        from neurotools.loading import load, load_data

        # Data are saved in separate folders
        subjs = os.listdir('data/ex1')
        subjs

        # Within each subject's directory is single data file
        # E.g., loading the first subjects data
        load('data/ex1/subj1/data.nii.gz')

        # Now we can load all subjects with the load_data method
        all_data = load_data(subjects=subjs,
                             template_path='data/ex1/SUBJECT/data.nii.gz')

        all_data.shape

    Additional arguments can of course be added allowing for more flexibility.
    Consider an additional case below:

    .. ipython:: python

        # We can add a mask, keeping only
        # the first and last data point
        mask = np.zeros(10)
        mask[0], mask[9] = 1, 1

        all_data = load_data(subjects=subjs,
                             template_path='data/ex1/SUBJECT/data.nii.gz',
                             mask=mask)

    '''

    # Error if both set
    if zero_as_nan and nan_as_zero:
        raise RuntimeError('zero_as_nan and nan_as_zero cannot both be True.')

    # Keep track of loading time
    start_time = time.time()
    
    # Either init _print or use passed value
    _print = _get_print(verbose, _print=_print)
    
    # Can pass mask as None, file loc, or numpy array
    mask, _ = _load_mask(mask)

    # Load data, with caching behavior
    data = _base_cache_load(load_args=(np.array(subjects), contrast, template_path,
                                       mask, index_slice, dtype),
                            load_func=_base_load_data,
                            cache_load_func=_cache_load_data,
                            cache_save_func=_cache_save_data,
                            cache_args=cache_args,
                            _print=_print,
                            no_cache_args={'n_jobs': n_jobs,
                                           'verbose': verbose})
                    
    _print('Loaded data with shape:', data.shape, 'in',
           time.time() - start_time,
           'seconds.', level=1)

    # Process zero_as_nan flag
    if zero_as_nan:
        _print(f'Setting {len(data[data == 0])} are 0 data points to NaN.')
        data[data == 0] = np.nan

    # Process nan_as_zero flag
    if nan_as_zero:
        _print(f'Setting {len(data[np.isnan(data)])} are NaN data points to 0.')
        data[np.isnan(data)] = 0

    return data

def reverse_mask_data(data, mask):
    '''This function is used to reverse the application of a mask
    to a data array.

    Parameters
    ------------
    data : :class:`numpy.ndarray` or list of
        Some data in which to transform from either
        a single dimensional :func:`array<numpy.array>` of data,
        or a 2D :func:`array<numpy.array>`
        of data with shape = subjs x data, back to the original shape
        of the data pre-masking. Instead of a 2D :func:`array<numpy.array>`, a list
        of single :func:`arrays<numpy.array>` can be passed as well.

    mask : loc or :class:`numpy.ndarray`
        The location of, or a :func:`array<numpy.array>` mask, where
        (1/True=keep, 0/False=discard) in which to reverse the
        application of. If the passed mask is either
        a :class:`Nifti1Image<nibabel.nifti1.Nifti1Image>` with an affine, or saved as a nifti,
        then as a part of the reverse mask process, the data
        will be converted into a :class:`Nifti1Image<nibabel.nifti1.Nifti1Image>` image.

    Returns
    --------
    reversed_data : :class:`numpy.ndarray`, :class:`Nifti1Image<nibabel.nifti1.Nifti1Image>` or list of
        In the case that the mask is passed in a format that
        provides an affine, data will be returned in :class:`Nifti1Image<nibabel.nifti1.Nifti1Image>`
        format. If instead the mask is passed as a :func:`array<numpy.array>` or
        location of a saved data array without affine information,
        then the data will be returned as a :func:`numpy array<numpy.array>`.

        If either a list of data :func:`arrays<numpy.array>` is passed, or data is passed
        as a 2D :func:`array<numpy.array>` with shape = subjs x data, then the returned
        reversed data will be a list with each element corresponding
        to each passed subject's reversed data.

    Examples
    ----------

    Consider the following brief example:

    .. ipython:: python

        import numpy as np
        from neurotools.loading import reverse_mask_data

        data = np.random.random((4, 2))
        mask = np.array([1, 0, 0, 1])
        
        reversed_data = reverse_mask_data(data, mask)
        reversed_data

        reversed_data[0].shape

    Or in the case where the mask is a :class:`Nifti1Image<nibabel.nifti1.Nifti1Image>`:

    .. ipython:: python
        
        import nibabel as nib

        # Convert mask to nifti
        mask = nib.Nifti1Image(mask, affine=np.eye(4))

        reversed_data = reverse_mask_data(data, mask)
        reversed_data
        
        # Closer look
        reversed_data[0].get_fdata()
        reversed_data[0].affine

    '''

    if mask is None:
        return None

    # Can pass mask as None, file loc, or numpy array
    mask, mask_affine = _load_mask(mask)

    # If multiple subjects
    if len(data.shape) == 2 or isinstance(data, list):
        
        projected = []
        for subj in data:
            projected.append(_project_single_subj(subj, mask, mask_affine))

        return projected

    # If single subject
    return _project_single_subj(data, mask, mask_affine)

def _project_single_subj(subj, mask, mask_affine):

    # Make copy of mask to fill with same dtype as subj's data
    proj_subj = mask.copy().astype(subj.dtype)
    proj_subj[mask == 1] = subj
    
    # Wrap in nifti image if needed
    if mask_affine is not None:
        proj_subj = nib.Nifti1Image(proj_subj, affine=mask_affine)

    return proj_subj

def get_overlap_subjects(df, template_path=None, contrast=None,
                         data_df=None, verbose=1, _print=None):
    '''Helper function to be used when working with template_path style saved
    data in order to compute an overlapping set of subjects between a dataframe
    and a template path, or alternatively between two dataframes.

    Parameters
    ----------
    df : :class:`pandas.DataFrame`
        A :class:`pandas.DataFrame` that is indexed by the names of the
        subjects in which to overlap.
        Where the names correspond in some way to the
        way the way the subject's data is saved.

    template_path : str, optional
        A str indicating the template form for how a single
        subjects data should be loaded (or in this case located),
        where SUBJECT will be replaced with that subjects name,
        and optionally CONTRAST will be replaced with the contrast name. 

        For example, to load subject X's contrast Y saved under
        'some_loc/X_Y.nii.gz'
        the template_path would be: 'some_loc/SUBJECT_CONTRAST.nii.gz'.

        Note that the use of a CONTRAST argument is optional.

        Note that if this parameter is passed, then data_df will
        be ignored!

        ::

            default = None

    contrast : str, optional
        The name of the contrast, used along with the template
        path to define where to load data.

        If passed :data:`None` then it is assumed that CONTRAST
        is not present in the template_path and will be ignored.

        Note that if this parameter is passed, then data_df will
        be ignored! This parameter is used only with the template_path option.

        ::

            default = None

    data_df : :class:`pandas.DataFrame` or :data:`None`, optional
        Optionally specify a :class:`pandas.DataFrame`, indexed by subject, in which
        to overlap with `df`, INSTEAD of a template_path and / or contrast set of arguments.
        Explicitly, if specifying a data_df, a template_path should not be passed!

        ::

            default = None

    verbose : int, optional
        By default this value is 1. This parameter
        controls the verbosity of this function.

        If -1, then no message at all will be printed.
        If 0, only warnings will be printed.
        If 1, general status updates will be printed.
        If >= 2, full verbosity will be enabled.

        ::

            default = 1

    '''

    # Either init _print or use passed value
    _print = _get_print(verbose, _print=_print)

    _print('Passed df with shape', df.shape, level=1)
    _print('Determining valid overlap subjects', level=1)
    
    # Only include subject if found as file
    if template_path is not None:
        all_subjects = [s for s in df.index if 
                        os.path.exists(_apply_template(subject=s,
                                                       template_path=template_path,
                                                       contrast=contrast))]

    # Unless computing overlap with data df
    else:
        all_subjects = [s for s in df.index if s in data_df.index]

    _print('Found', len(all_subjects), 'subjects with data.', level=1)
    
    # Print missing subjects if high enough verbose
    missing_subjects = [s for s in df.index if s not in all_subjects]
    _print('Missing:', missing_subjects, level=2)

    return all_subjects




import numpy as np
import random
import pandas as pd
from sklearn.model_selection import train_test_split
from joblib import Parallel, delayed
from sklearn.model_selection import GroupShuffleSplit
from scipy.stats import pearsonr, ttest_1samp
from statsmodels.stats.multitest import fdrcorrection
from ..stats.basic import get_cohens, get_resid, fast_corr
from ..loading.funcs import load_data, reverse_mask_data, get_overlap_subjects
from ..misc.print import _get_print


def get_non_nan_overlap_mask(c1, c2):
    return ~np.isnan(c1) & ~np.isnan(c2)

def get_corr_size_n(covars=None, data=None, resid=None,
                    base_map=None, proc_covars_func=None,
                    perf_series=None,
                    n=None, thresh=None):

    # Set a random seed based on another random seed (useful in multiproc context)
    np.random.seed(random.randint(1, 10000000))

    # Select a random group of size n
    if data is None and resid is not None:
        index = np.arange(0, len(resid))
    else: 
        index = np.arange(0, len(data))

    choices = np.random.choice(index, n, replace=False)

    # If split == every, go through full proc
    if data is not None and covars is not None:
        
        if perf_series is None:
            ps = None
        else:
            ps = perf_series.iloc[choices].copy()
        
        compare_map = get_proc_map(covars.iloc[choices].copy(), data[choices],
                                   proc_covars_func, perf_series=ps)

    # Otherwise, just calculate based on subset of passed resid
    else:

        if perf_series is None:
            compare_map = get_cohens(resid[choices])
        else:
            compare_map = fast_corr(resid[choices],
                                    np.array(perf_series.iloc[choices].copy()))

    # In the case that the calculated map has any NaN
    # calculate a mask of only non NaN values in either compare_map
    valid = get_non_nan_overlap_mask(compare_map, base_map)

    if thresh is not None:
        valid = valid & (np.abs(base_map) > thresh)

    # Calculate the correlation only for the overlap
    corr = np.corrcoef(compare_map[valid], base_map[valid])[0][1]
    corr, p_value = pearsonr(compare_map[valid], base_map[valid])

    return corr, p_value

def get_corrs(x_labels=None, covars=None, data=None, resid=None,
              base_map=None, proc_covars_func=None, perf_series=None,
              thresh=None, _print=print):

    corrs = []
    p_values = []

    for n in x_labels:
        corr, p_value =\
            get_corr_size_n(covars=covars, data=data, resid=resid, 
                            base_map=base_map, proc_covars_func=proc_covars_func,
                            perf_series=perf_series,
                            n=n, thresh=thresh)
        
        _print('Corr size n=', n, '=', corr, 'p_value=', p_value, level=2)
        corrs.append(corr)
        p_values.append(p_value)

    _print('Finished repeat!', level=1)

    return corrs, p_values

def get_proc_map(covars, data, proc_covars_func, perf_series=None):

     # Proc separate if passed
    if proc_covars_func is not None:
        covars = proc_covars_func(covars)
    
    # Residualize
    resid = get_resid(covars, data)

    # Generate cohens if no perf_series
    if perf_series is None:
        return get_cohens(resid)

    # Otherwise, generate correlation w/ perf_series
    else:
        return fast_corr(resid, np.array(perf_series))

def rely(proc_type, covars, data, base_map, proc_covars_func,
         perf_series, x_labels, n_repeats, thresh, n_jobs, _print):

    if proc_type == 'split':
        _print('Starting Reliability Test with proc_type = "split"')
        
        # Proc separate if passed
        if proc_covars_func is not None:
            covars = proc_covars_func(covars)
            _print('Applied proc_covars_func on each covars df seperately')

        # Residualize
        _print('Generate Residualized Data')
        resid = get_resid(covars, data)

        # Set None
        covars, data, proc_covars_func = None, None, None

    elif proc_type == 'every':
        _print('Starting Reliability Test with proc_type = "every"')

        # Set resid None
        resid = None

    else:
        raise RuntimeError('proc_type must be "split" or "every"')

    _print('Starting Reliability Test')
    output = Parallel(n_jobs=n_jobs)(
            delayed(get_corrs)(x_labels=x_labels,
                               covars=covars,
                               data=data,
                               resid=resid,
                               base_map=base_map,
                               proc_covars_func=proc_covars_func,
                               perf_series=perf_series,
                               thresh=thresh,
                               _print=_print) for _ in range(n_repeats))

    all_corrs = [o[0] for o in output]
    all_p_values = [o[1] for o in output]
         
    return all_corrs, all_p_values

def _test_split(all_subjects, stratify, groups, split_random_state):
    
    sorted_subjects = np.array(sorted(all_subjects))

    if groups is not None and stratify is not None:
        raise RuntimeError('groups and stratify cannot both be specified.')

    # Proc the stratify series
    if stratify is not None:
        stratify_vals = stratify.loc[sorted_subjects]
    else:
        stratify_vals = None

    # Proc the group series
    if groups is not None:
        groups_vals = groups.loc[sorted_subjects]
    else:
        groups_vals = None
        
    if groups is not None:
        
        splitter = GroupShuffleSplit(n_splits=1, test_size=.5,
                                     random_state=split_random_state)
        inds = splitter.split(sorted_subjects, groups=groups_vals)
        inds = [*inds]

        g1_subjects, g2_subjects = sorted_subjects[inds[0][0]], sorted_subjects[inds[0][1]]
    
    else:
    
        # Compute the train test split on the sorted subjects (for reproducibility)
        g1_subjects, g2_subjects = train_test_split(sorted_subjects,
                                                    test_size=.5,
                                                    random_state=split_random_state,
                                                    stratify=stratify_vals)

    return g1_subjects, g2_subjects

                                
def run_rely(covars_df, data_df=None,
             contrast=None, template_path=None, mask=None,
             index_slice=None, stratify=None, groups=None,
             proc_covars_func=None,
             perf_series=None, proc_type='split',
             thresh=None, min_size=5,
             max_size=1000, every=1, n_repeats=100,
             n_jobs=1, split_random_state=2, verbose=1):
    ''' Function for computing a basic metric of reliability.

    Notes on NaN's:
    - NaN's may not be passed in the covars_df, stratify or perf_series.

    - If there are any NaN's in the data, when residualizing each voxel/vertex,
    that subject will be excluded from calculating that voxel/vertex's residual,
    but will still be included in every case where they have a non-nan value.
    Essentially the NaN value will propegate into the residualized data.
    Likewise, when calculating the cohen's map, any nan values will simply
    be skipped when calculating the mean and std. 

    - If there are any NaN's in either the group to compares cohens map,
    or the base cohen's map, then those voxels / vertex will be excluded
    when calculating the correlation coef. between the two maps. NaN's
    will show up here if the standard deviation for that feature
    across subjects is 0.
    
    Parameters
    ----------
    covars_df : pandas DataFrame
        A pandas dataframe containing any co-variates in which
        the loaded data should be residualized by.

        Note: The dataframe must be indexed by subject name!

    data_df : pandas DataFrame or None
        Dataframe indexed by subject name that contains
        the data to be used. Note if provided then
        this will replace any arguments passed to
        contrast, template_path, mask and index_slice.

    contrast : str
        The name of the contrast, used along with the template
        path to define where to load data.

        This parameter is ignored if data_df is not None.

    template_path : str
        A str indicating the template form for how a single
        subjects data should be loaded, where SUBJECT will be
        replaced with that subjects name, and CONTRAST will
        be replaced with the contrast name.

        For example, so load subject X's contrast Y saved at:
        some_loc/X_Y.nii.gz.
        
        You would pass:
        some_loc/SUBJECT_CONTRAST.nii.gz

        As the template path.

        This parameter is ignored if data_df is not None.

    mask : str, numpy array or None
        After data is loaded, it can optionally be
        masked. By default, this parameter is set to None.
        If None, then the subjects data will be flattened.
        If passed a str, it will be assumed to be the location of a mask
        in which to load, which will then use nibabel's load function
        and then get_fdata() to extract a binary mask, where the shape
        of the mask should match the data and also entries set to True or
        1, indicate that that value be kept when loading data.
        Lastly, a numpy array, where 1 == a value should be kept, with
        likewise the same shape as the data to load can be passed here.

        This parameter is ignored if data_df is not None.

    index_slice : slices, tuple of slices, or None
        You may optional pass index slicing here. This will be applied
        before the mask. The benefit of using index slicing over a mask
        is that in the case where say your data is of shape 5 x 20 on
        the disk, but you only need the data in the first 20, e.g.,
        so with traditional slicing that is [0], in this case by using the
        mask you have to load the full data from the disk, then slice it.
        But if you pass slice(0) here, then this accomplishes the same thing,
        but only loads the requested slice from the disk.

        You must use slice for complex slicing, in the example above you could just
        pass (0) or slice(0), but if you wanted something more complex, e.g. passing
        something like my_array[1:5:2, ::3] you should pass
        (slice(1,5,2), slice(None,None,3)) here.
        
        By default this is None.

        This parameter is ignored if data_df is not None.

    stratify : None or pandas Series
        By default this is None. If passed a series though,
        then this series is expected to be indexed by subject,
        with the same subjects as the passed covars_df. This series
        of values will then be passed to the train_test_split function,
        and will specify that stratifying behavior is requested.

        Stratifying behavior will be applied for only the main
        split, and will ensure or rather try to ensure that the
        same distribution of class values will be present in each split.

        If this is passed as a series, then groups must be None.

    groups : None or pandas Series
        By default this is None, if passed a series, then this
        series is expected to be indexed by subject, with
        the same subjects as the passed covars_df. This series
        will then be used to define the main train_test_split,
        but will perform it according to group preserving behavior,
        where members of the same group (as passed here) will
        be preserved within the same split.

        If this is passed as a series, then stratify must be None.

    proc_covars_func : None, function
        By default, this is set to None. Alternatively, you
        may pass a function in which the first positional argument accepts
        a subset of the covars_df, and then returns a processed version of
        the covar df. This is useful for preforming pre-processing on the
        covars_df separately for each group of subjects.

    perf_series : None or pandas Series
        By default None. If not None, then this should
        be a series indexed by subject id. The cohen's
        will not be calculated anymore, instead the reliability
        for the residualized data as correlated with this
        series will be computed instead.

    proc_type : 'split' or 'every'
        This defines the behavior of the reliability test.
        In the first case, 'split', the covars df will be processed
        seperately across each of the two main groups (if proc_covars_func is None),
        and also each groups residualized data computed on each full half.
        
        Alternatively, if 'every' is passed, then the gold standard group will still be processed
        all together, but the comparison group will be processed and likewise
        residualized seperately for each comparison group! Warning: this
        can take longer to compute.

    thresh : float or None
        By default, this is set to None. This value if not none
        indicates an absolute value threshold in which only voxels / vertex
        in the comparison map above this threshold should be used
        to calculate the correlation between the random group and this
        base map.

    min_size : int
        By default this is set to 5. This is the starting
        group size to compare with the split half sample.

        Where random groups sizes are defined by:
        x_labels = list(range(min_size, max_size, every))

    max_size : int
        By default this is set to 1000. This is the maximum
        size to compare with the split half sample.

        Where random groups sizes are defined by:
        x_labels = list(range(min_size, max_size, every))

    every : int
        By default this value is 1, it determines
        jumps to the groups size.

        Where random groups sizes are defined by:
        x_labels = list(range(min_size, max_size, every))

    n_repeats : int
        By default = 100. This value controls
        how many times each of the group sizes should
        be evaluated with a different random group of that size.

    n_jobs : int
        The number of jobs to try and use for loading and the rely test.

    split_random_state : int
        By default 2. Can pass different int's.
        This is the random state for the random tr test split.

    verbose : int
        By default this value is 1. This parameter
        controls the verbosity of this function.

        If -1, then no message at all will be printed.
        If 0, only warnings will be printed.
        If 1, general status updates will be printed.
        If >= 2, full verbosity will be enabled.
    '''
    
    # Get print
    _print = _get_print(verbose=verbose)
    
    # Get overlapping subjects
    all_subjects = get_overlap_subjects(df=covars_df, template_path=template_path,
                                        contrast=contrast,
                                        data_df=data_df, _print=_print)

    _print('Performing group split, w/ stratify =', stratify is not None,
           ', w/ groups =', groups is not None, ', '
           'random_state =', split_random_state)

    # Apply split
    g1_subjects, g2_subjects = _test_split(all_subjects=all_subjects,
                                           stratify=stratify,
                                           groups=groups,
                                           split_random_state=split_random_state)
    assert len(set(g1_subjects).intersection(set(g2_subjects))) == 0, "Issue Splitting Groups!"
    _print('len(group1) =', len(g1_subjects), 'len(group2) =', len(g2_subjects))

    if data_df is None:

        # Load the data from files
        _print('Loading Group 1 Data')
        d1 = load_data(subjects=g1_subjects,
                      contrast=contrast,
                      template_path=template_path,
                      mask=mask,
                      index_slice=index_slice,
                      n_jobs=n_jobs,
                      _print=_print)

        _print('Loading Group 2 Data')
        d2  = load_data(subjects=g2_subjects,
                       contrast=contrast,
                       template_path=template_path,
                       mask=mask,
                       index_slice=index_slice,
                       n_jobs=n_jobs,
                       _print=_print)
       
    else:

        # Load data from the data df
        _print('Loading Group 1 and 2 Data from data_df')
        d1 = np.array(data_df.loc[g1_subjects])
        d2 = np.array(data_df.loc[g2_subjects])

     # Assign variable to the co-variates per group
    c1 = covars_df.loc[g1_subjects].copy()
    c2 = covars_df.loc[g2_subjects].copy()

    # If perf series is passed, make sure it is set to the right subjects
    if perf_series is not None:
        p1 = perf_series.loc[g1_subjects].copy()
        p2 = perf_series.loc[g2_subjects].copy()
    else:
        p1, p2 = None, None

    # Generate x_labels
    x_labels = list(range(min_size, max_size, every))

    # Get base map, cohens or perf corr
    _print('Generate Base/Comparison Map')
    base_map = get_proc_map(c1, d1, proc_covars_func, perf_series=p1)

    # Print out base thresh info, if thresh passed
    if thresh is not None:
        above_thresh=np.sum(np.abs(base_map) > thresh)
        _print(above_thresh, 'above passed pass thresh=', thresh)

    # Run rely separate based on proc_type
    all_corrs, all_p_values = rely(
        proc_type=proc_type, covars=c2, data=d2,
        base_map=base_map, proc_covars_func=proc_covars_func,
        perf_series=p2, x_labels=x_labels, n_repeats=n_repeats,
        thresh=thresh, n_jobs=n_jobs, _print=_print)
  
    # Convert to array and means by repeat
    all_corrs = np.array(all_corrs)
    corr_means = np.mean(all_corrs, axis=0)
    corr_stds = np.mean(all_corrs, axis=0)

    # Same with p_values
    all_p_values = np.array(all_p_values)
    p_value_means = np.mean(all_p_values, axis=0)
    p_value_stds = np.mean(all_p_values, axis=0)

    return x_labels, corr_means, corr_stds, p_value_means, p_value_stds

def load_resid_data(covars_df, contrast, template_path, mask=None,
                    index_slice=None, resid=True, n_jobs=1, verbose=1):
    ''' Loading residualized data. Or non-residualized... 

    Note: Unlike run_rely, there is not proc_covars_func, as
    this function assumes just one group to load, therefore
    the passed covars_df should be suitably proc'ed before passing
    here.
    
    Parameters
    ----------
    covars_df : pandas DataFrame
        A pandas dataframe containing any co-variates in which
        the loaded data should be residualized by.

        Note: The dataframe must be indexed by subject name!

    contrast : str
        The name of the contrast, used along with the template
        path to define where to load data.

    template_path : str
        A str indicating the template form for how a single
        subjects data should be loaded, where SUBJECT will be
        replaced with that subjects name, and CONTRAST will
        be replaced with the contrast name.

        For example, so load subject X's contrast Y saved at:
        some_loc/X_Y.nii.gz.
        
        You would pass:
        some_loc/SUBJECT_CONTRAST.nii.gz

        As the template path.

    mask : str, numpy array or None
        After data is loaded, it can optionally be
        masked. By default, this parameter is set to None.
        If None, then the subjects data will be flattened.
        If passed a str, it will be assumed to be the location of a mask
        in which to load, which will then use nibabel's load function
        and then get_fdata() to extract a binary mask, where the shape
        of the mask should match the data and also entries set to True or
        1, indicate that that value be kept when loading data.
        Lastly, a numpy array, where 1 == a value should be kept, with
        likewise the same shape as the data to load can be passed here.

    index_slice : slices, tuple of slices, or None
        You may optional pass index slicing here. This will be applied
        before the mask. The benefit of using index slicing over a mask
        is that in the case where say your data is of shape 5 x 20 on
        the disk, but you only need the data in the first 20, e.g.,
        so with traditional slicing that is [0], in this case by using the
        mask you have to load the full data from the disk, then slice it.
        But if you pass slice(0) here, then this accomplishes the same thing,
        but only loads the requested slice from the disk.

        You must use slice for complex slicing, in the example above you could just
        pass (0) or slice(0), but if you wanted something more complex, e.g. passing
        something like my_array[1:5:2, ::3] you should pass
        (slice(1,5,2), slice(None,None,3)) here.
        
        By default this is None.

    resid : bool
        If the data should be residualized or not. True by default
        for yes, if False, then raw data will be returned.

    n_jobs : int
        The number of jobs to try and use for loading and the rely test.

    verbose : int
        By default this value is 1. This parameter
        controls the verbosity of this function.

        If -1, then no message at all will be printed.
        If 0, only warnings will be printed.
        If 1, general status updates will be printed.
        If >= 2, full verbosity will be enabled.
    '''

    # Get print
    _print = _get_print(verbose=verbose)

    # Check inputs
    if not isinstance(covars_df, pd.DataFrame):
        raise RuntimeError('covars_df must be a DataFrame.')

    if template_path is not None:
        if not isinstance(template_path, str):
            raise RuntimeError('template_path must be a str.')
    
    # Get overlapping subjects
    all_subjects = get_overlap_subjects(df=covars_df,
                                        template_path=template_path,
                                        contrast=contrast,
                                        _print=_print)

    # Set covars to just the subjects found
    covars = covars_df.loc[all_subjects].copy()

    # Load all data
    _print('Loading Data', level=1)
    data = load_data(subjects=all_subjects,
                    contrast=contrast,
                    template_path=template_path,
                    mask=mask,
                    index_slice=index_slice,
                    n_jobs=n_jobs,
                    _print=_print)

    if resid:

        _print('Residualizing Data', level=1)
        resid = get_resid(covars, data)

        return all_subjects, resid

    else:

        _print('Returning Raw Data', level=1)
        return all_subjects, data

def run_1samp_summary(covars_df, contrast,
                      template_path, popmean=0,
                      alpha=.05,
                      mask=None, reverse_mask=True,
                      index_slice=None,
                      n_jobs=1, verbose=1):
    '''
    For the t-test uses:
    https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.ttest_1samp.html

    For p-value correction uses:
    https://www.statsmodels.org/stable/generated/statsmodels.stats.multitest.fdrcorrection.html

    Parameters
    -----------
    covars_df : pandas DataFrame
        A pandas dataframe containing any co-variates in which
        the loaded data should be residualized by.

        Note: The dataframe must be indexed by subject name!

    contrast : str
        The name of the contrast, used along with the template
        path to define where to load data.

    template_path : str
        A str indicating the template form for how a single
        subjects data should be loaded, where SUBJECT will be
        replaced with that subjects name, and CONTRAST will
        be replaced with the contrast name.

        For example, so load subject X's contrast Y saved at:
        some_loc/X_Y.nii.gz.
        
        You would pass:
        some_loc/SUBJECT_CONTRAST.nii.gz

        As the template path.

    popmean : float or array_like, optional
        Expected value in null hypothesis. If array_like,
        then it must have the same shape as a excluding the axis dimension.

        default = 0

    alpha : float
        The error rate in the fdr correction.

        covars_df : pandas DataFrame
        A pandas dataframe containing any co-variates in which
        the loaded data should be residualized by.

        Note: The dataframe must be indexed by subject name!

    mask : str, numpy array or None
        After data is loaded, it can optionally be
        masked. By default, this parameter is set to None.
        If None, then the subjects data will be flattened.
        If passed a str, it will be assumed to be the location of a mask
        in which to load, which will then use nibabel's load function
        and then get_fdata() to extract a binary mask, where the shape
        of the mask should match the data and also entries set to True or
        1, indicate that that value be kept when loading data.
        Lastly, a numpy array, where 1 == a value should be kept, with
        likewise the same shape as the data to load can be passed here.

    reverse_mask : bool, optional
        Can optionally reverse the masking, assuming that no index_slice is
        passed! If mask is passed as array, then will project data
        back as array. If mask is passed as location to nii, then data will
        be put back as a nifti object.

    index_slice : slices, tuple of slices, or None
        You may optional pass index slicing here. This will be applied
        before the mask. The benefit of using index slicing over a mask
        is that in the case where say your data is of shape 5 x 20 on
        the disk, but you only need the data in the first 20, e.g.,
        so with traditional slicing that is [0], in this case by using the
        mask you have to load the full data from the disk, then slice it.
        But if you pass slice(0) here, then this accomplishes the same thing,
        but only loads the requested slice from the disk.

        You must use slice for complex slicing, in the example above you could just
        pass (0) or slice(0), but if you wanted something more complex, e.g. passing
        something like my_array[1:5:2, ::3] you should pass
        (slice(1,5,2), slice(None,None,3)) here.
        
        By default this is None.

    n_jobs : int
        The number of jobs to try and use for loading and the rely test.

    verbose : int
        By default this value is 1. This parameter
        controls the verbosity of this function.

        If -1, then no message at all will be printed.
        If 0, only warnings will be printed.
        If 1, general status updates will be printed.
        If >= 2, full verbosity will be enabled.
    '''

    # Load resid data
    all_subjects, data =\
        load_resid_data(covars_df=covars_df, contrast=contrast,
                        template_path=template_path, mask=mask,
                        index_slice=index_slice, resid=True, n_jobs=n_jobs, verbose=verbose)

    results = {}
    
    # Run ttest
    ttest_result = ttest_1samp(data, popmean=popmean, nan_policy='omit')
    results['statistic'] = ttest_result.statistic
    results['uncorrected_p_values'] = ttest_result.pvalue

    # Run FDR
    results['corrected_p_values'] = fdrcorrection(ttest_result.pvalue, alpha=alpha)[1]
    
    # Gen cohen's map
    results['cohens'] = get_cohens(data)
    
    # If reverse_mask, reverse
    if reverse_mask:
        for key in results:
            results[key] = reverse_mask_data(results[key], mask=mask)

    results['subjects'] = all_subjects
    return results
  



import os
from ..misc.text import readable_size_to_bytes
from pathlib import Path

from joblib.hashing import hash as joblib_hash

def _get_cache_dr(cache_dr, load_func):

    # Process default
    if cache_dr == 'default':
        func_name = load_func.__name__
        default_cache_dr = os.path.join(os.path.expanduser("~"), 'neurotools_cache')
        cache_dr = os.path.join(default_cache_dr, func_name)

    # Give error if exists and not directory
    if os.path.exists(cache_dr) and not os.path.isdir(cache_dr):
        raise RuntimeError(f'Passed cache_dr: {cache_dr} already exists and is not a directory!')

    # Make sure exists if doesn't exist
    os.makedirs(cache_dr, exist_ok=True)

    return cache_dr

def _get_load_hash_str(load_args, use_base_name):

    to_cache_args = []
    for arg in load_args:

        # None case
        if arg is None:
            to_cache_args.append(None)

        # If str, check if location, if location use
        # basename instead of absolute path
        elif use_base_name and isinstance(arg, str) and os.path.exists(arg):
            to_cache_args.append(os.path.basename(arg))

        # If list, sort
        elif isinstance(arg, list):
            to_cache_args.append(sorted(arg))

        # Otherwise add directly
        else:
            to_cache_args.append(arg)

    # Return joblib hash
    return joblib_hash(to_cache_args)

def _get_load_cache_loc(load_args, cache_dr, use_base_name):

    # Otherwise, caching behavior is expected
    # Get unique hash for these arguments - with tolerance to order and
    # and a few other things.
    hash_str = _get_load_hash_str(load_args, use_base_name)

    # Next, we want to be able to handle arbitrary file ext.
    # an existing cache_loc might be saved under
    existing_cached = os.listdir(cache_dr)

    # Get just hashes
    existing_hashes = [file.split('.')[0] for file in existing_cached]

    # Check if hash str is an existing hash
    if hash_str in existing_hashes:

        # If exists, replace hash_str with full
        # hash + file ext.
        hash_str = existing_cached[existing_hashes.index(hash_str)]

    # Set full path
    cache_loc = os.path.join(cache_dr, hash_str)

    return cache_loc

def _unpack_cache_args(cache_args):
    '''Unpack possible args, otherwise set to default'''

    if 'cache_dr' in cache_args:
        cache_dr = cache_args['cache_dr']
    else:
        cache_dr = 'default'

    if 'cache_max_sz' in cache_args:
        cache_max_sz = cache_args['cache_max_sz']
    else:
        cache_max_sz = '30G'

    if 'use_base_name' in cache_args:
        use_base_name = cache_args['use_base_name']
    else:
        use_base_name = True

    # TODO warn if any others passed and not used

    return cache_dr, cache_max_sz, use_base_name

def _base_cache_load(load_args, load_func,
                     cache_load_func, cache_save_func,
                     cache_args, _print, no_cache_args=None):

    if no_cache_args is None:
        no_cache_args = {}

    # Unpack cache args
    cache_dr, cache_max_sz, use_base_name = _unpack_cache_args(cache_args)

    # If no caching, base case, just load as normal
    if cache_dr is None:
        _print(f'No cache_dr specified, loading directly', level=1)
        return load_func(*load_args, **no_cache_args)

    # Make sure cache_dr arg
    cache_dr = _get_cache_dr(cache_dr, load_func)

    # Get cache loc
    cache_loc = _get_load_cache_loc(load_args, cache_dr, use_base_name)

    # If exists, load from saved
    if os.path.exists(cache_loc):
        _print(f'Loading from cache_loc: {cache_loc}', level=1) 

        # Load with correct function
        data = cache_load_func(cache_loc, load_args)

        # Make sure to updating last modified of file to now
        Path(cache_loc).touch()

    # If doesn't yet exist, load like normal
    else:
        _print(f'No existing cache found, loading directly', level=1)

        # Base load
        data = load_func(*load_args, **no_cache_args)

        # Save with special cache func
        cache_save_func(data=data, cache_loc=cache_loc)

        _print(f'Saving loaded data to cache_loc: {cache_loc}', level=1)

    # Before returning - check the cache_max_sz argument
    # clearing any files over the limit
    _check_cache_sz_limit(cache_dr, cache_max_sz, _print=_print)

    return data

def _check_cache_sz_limit(cache_dr, cache_max_sz, _print):

    _print('Checking cache size limit', level=2)

    # Get full file paths of all cached
    all_cached = [os.path.join(cache_dr, f)
                  for f in os.listdir(cache_dr)]

    # Get size of current directory
    size = sum(os.path.getsize(f) for f in all_cached)

    # Make sure cache_max_sz as bytes
    cache_max_sz = readable_size_to_bytes(cache_max_sz)

    _print(f'Current saved size is: {size} / {cache_max_sz} (cache_max_sz)', level=1)

    # If over the current limit
    if size > cache_max_sz:

        # Sort all cached files from oldest to newest, based
        # on last modified.
        old_to_new = sorted(all_cached, key=os.path.getctime)

        # Delete in order from old to new
        # until under the limit
        removed, n = 0, 0
        while size - removed > cache_max_sz:

            # Select next to remove
            to_remove = old_to_new[n]
            
            # Update counters
            n += 1
            removed += os.path.getsize(to_remove)

            # Remove cached file
            os.remove(to_remove)

            _print(f'Removed cached file at: {to_remove}', level=2)


import numpy as np
import pandas as pd
from sklearn.utils import check_random_state
from sklearn.preprocessing import LabelEncoder
import warnings

def _within_grp_permute(blocks, rng):
    '''If here, means performing permutations within each
    unique group. By default this should be called first?'''
    
    # If reached the index layer - stop
    if blocks.shape[1] == 1:
        return blocks
    
    # Get all unique groups
    col = blocks[:, 0]
    unique_grps = np.unique(col)
    
    # Process each group separate
    for grp in unique_grps:
        
        # Get index of members of this group
        grp_inds = np.where(col == grp)[0]
        
        # Neg case, within grp permute
        if grp < 0:
            blocks[grp_inds, 1:] = _within_grp_permute(blocks[grp_inds, 1:], rng)
        
        # Pos case is swapping groups
        else:
            blocks[grp_inds, 1:] = _between_grp_permute(blocks[grp_inds, 1:], rng)
            
    return blocks
            

def _between_grp_permute(blocks, rng):
    '''Permute between groups of the same size'''
    
    # Get all unique groups
    col = blocks[:, 0]
    unique_grps, cnts = np.unique(col, return_counts=True)
    
    # Fill in new_blocks with permuted copy
    new_blocks = blocks.copy()
    
    # For each unique size of group, permute with other groups
    for grp_size in np.unique(cnts):
        
        # Get a permuted order for all valid groups
        grps_of_size = unique_grps[np.where(cnts==grp_size)]
        permuted_grps = rng.permutation(grps_of_size)
        
        # Apply the permuted order
        original_inds = [np.where(col == grp) for grp in grps_of_size]
        new_inds = [np.where(col == grp) for grp in permuted_grps]
        
        for o_inds, n_inds in zip(original_inds, new_inds):
            new_blocks[n_inds] = blocks[o_inds]
            
    # Still need to apply within grp - if any, to next level of data
    # Return the results of that operation
    return _within_grp_permute(new_blocks, rng)

def _proc_block_input(in_blocks, within_grp=True):

    # If passed as df, convert to array
    if isinstance(in_blocks, (pd.DataFrame, pd.Series)):
        in_blocks = np.array(in_blocks)

    # Passed 1D array case - expand
    if len(in_blocks.shape) == 1:
        in_blocks = np.expand_dims(in_blocks, axis=1)

    # Get outer col to add to left side
    ones = np.ones((len(in_blocks), 1), dtype='int')
    if within_grp:
        ones *= -1
    
    # Add index as column to blocks - to keep track of permutations
    indx = np.expand_dims(np.arange(in_blocks.shape[0], dtype='int'), 1)

    # Add index and ones cols and make sure blocks are int
    blocks = np.hstack([ones, in_blocks, indx]).astype('int')

    # Generate lexsort order from left to right,
    # from in_blocks and apply. Basically we need to apply this
    # ordering in order to ensure that any whole-block shuffling that occurs,
    # occurs such that the order is preserved.
    lex_sort_order = np.lexsort([in_blocks[:, -i] for i in range(in_blocks.shape[1])])
    blocks = blocks[lex_sort_order]

    # Reverse initial lex sort, and return that along with the blocks
    rev_lex_sort_order =\
        [np.where(lex_sort_order == i)[0][0] for i in range(len(lex_sort_order))]

    return blocks, rev_lex_sort_order


def permute_blocks(in_blocks, rng, within_grp=True):
    '''Function to perform permutations according to a PALM style
    Exchangeability Block structure.

    The sorting by lex order, then reversing is
    to make sure inter block swaps are done in a valid way according
    to any subtypes
    '''

    # Check for index attribute
    index = None
    if hasattr(in_blocks, 'index'):
        index = in_blocks.index

    # Prep blocks for permutations - adding ones and index cols
    # Then sorting, such that we can make assumptions
    # about the ordering of whole block swaps
    blocks, rev_lex_sort_order = _proc_block_input(in_blocks, within_grp=within_grp)

    # Start always as within grp on the outer layer of exchange blocks
    # the within_grp arg controls the behavior here.
    permuted_blocks = _within_grp_permute(blocks, rng)

    # Apply the reverse of the lexical sorting
    permuted_blocks = permuted_blocks[rev_lex_sort_order]
    
    # Only need to return the new permuted index
    permuted_indx = permuted_blocks[:, -1]

    # If passed an object originally w/ index, return
    # the permuted order of the actual index names
    if index is not None:
        return index[permuted_indx], index

    # Otherwise, return the permuted indx as arranged integer index
    return permuted_indx, index

def block_permutation(x, blocks, random_state=None, within_grp=True):
    '''
    Randomly permute a sequence, or return a permuted range.
    If x is a multi-dimensional array, it is only shuffled along its first index.
    Performs permutations according to passed blocks, where blocks
    are specified with PALM style Exchangeability Blocks.
    '''
    
    # Proc. random state
    rng = check_random_state(random_state)

    # Permute blocks once - in addition  to return permuted index
    # we also need the block index, in case order was different between
    # x and blocks and we need to use index
    permuted_indx, block_index = permute_blocks(blocks, rng, within_grp=within_grp)

    # Series case
    if isinstance(x, pd.Series):
        
        # Save original type
        original_type = x.dtype

        # Replace values w/ array of original
        x.loc[block_index] = np.array(x.loc[permuted_indx])

        # Cast back  to original type
        x = x.astype(original_type)

        return x

    # Dataframe case
    elif isinstance(x, pd.DataFrame):

        # Save original types
        original_types = x.dtypes

        # Replace values w/ array of original
        x.loc[block_index, :] = np.array(x.loc[permuted_indx])
        
        # Make sure each col's dtype stays the same
        for dtype, col in zip(original_types, x.columns):
            x[col] = x[col].astype(dtype)

        return x

    # Check for potential unexpected case and warn
    if block_index is not None:
        warnings.warn('Passed blocks had index, but object to permute does not.')


    # Base case, return original sequence
    # in new permuted order
    return x[permuted_indx]


def get_auto_vg(in_blocks, within_grp=True):
    '''The idea with the variance groups is that
    any subjects which can swap with each other should
    be in the same variance group
    but any that cannot, should be in different ones.'''

    # Prep blocks for determine vg
    blocks, _ = _proc_block_input(in_blocks, within_grp=within_grp)
    
    # Init vg's as empty array
    vgs = np.zeros(len(blocks), dtype='int')
    can_swap = {i: set() for i in range(len(blocks))}
    
    def set_vg_val(ind, vg):
        
        # If ind is array-like
        if hasattr(ind, '__iter__'):
            
            # If both are array-like
            if hasattr(vg, '__iter__'):
                for i, v in zip(ind, vg):
                    set_vg_val(i, v)
            
            # If just ind is array like
            else:
                for i in ind:
                    set_vg_val(i, vg)
        
        # Set single value case
        else:
            
            # When trying to set a new
            # value, we need to check
            # to see if the ind being
            # set has already been set by
            # one of its can swaps
            # if it has be set by a can swap
            # use that value instead of 
            # the one being set
            for s_i in can_swap[ind]:
                if vgs[s_i] != 0:
                    vgs[ind] = vgs[s_i]
                    return
            
            # Otherwise, set as is
            vgs[ind] = vg

    def within_grp_case(blocks, vg):
        
        # Get current column
        col = blocks[:, 0]

        # If at final index layer, set to passed vg
        if blocks.shape[1] == 1:
            set_vg_val(col, vg)
            return 

        # Get all unique groups
        unique_grps = np.unique(col)

        # For each group in the within group
        # case, the idea is that they should
        # get a new unique vg
        for grp in unique_grps:

            # Get index of members of this group
            grp_inds = np.where(col == grp)[0]
            
            # Get next highest vg
            new_vg = np.max(vgs) + 1

            # Neg, within group case
            if grp < 0:
                within_grp_case(blocks[grp_inds, 1:], vg=new_vg)

            # Pos case is swapping groups
            else:
                between_grp_case(blocks[grp_inds, 1:], vg=new_vg)

        return blocks
    
    def between_grp_case(blocks, vg):
        
        # Get current column
        col = blocks[:, 0]

        # If at final index layer, set to passed vg
        if blocks.shape[1] == 1:
            set_vg_val(col, vg)
            return 
        
        # Okay for the between grp case, what
        # we want to do is assign the same variance
        # group sequentially to all groups of the same size.
        
        # Get all unique groups and their sizes
        unique_grps, cnts = np.unique(col, return_counts=True)
        
        # For each unique size of group, keep track of
        # unique in pass along
        cnt, pass_along = vg, np.zeros(len(col))
        for grp_size in np.unique(cnts):

            # Get a permuted order for all valid groups
            grps_of_size = unique_grps[np.where(cnts==grp_size)]
            grp_inds = np.array([np.where(col == grp)[0] for grp in grps_of_size])
            
            # Set each sequentially
            for inds in grp_inds.T:
                pass_along[inds] = cnt
                cnt += 1
                
        # Get next_col
        next_col = blocks[:, 1]
                
        # If next layer is the index, end case
        if blocks.shape[1] == 2:
            set_vg_val(next_col, pass_along)
            return
        
        # Get unique groups as combo between columns
        temp_col = np.stack([next_col, pass_along], axis=1)
        unique_grps = np.unique(temp_col, axis=0)
        
        # Proc each unique combo group simmilar to within_grp_case
        # Basically the idea is that, every member
        # in the same group here should be in the same
        # variance group as they can swap.
        # We just need to consider the extra cases
        # where based on the values of the next col,
        # there might be further overlapping groups
        for grp in unique_grps:
            
            # Get grp inds as binary array
            grp_inds = np.all(temp_col == grp, axis=1)
            
            swaps = set(blocks[grp_inds, -1])
            for i in swaps:
                can_swap[i] = swaps
                
        # After values are set in can swap,
        # we can process the groups as is with
        # within group case - doesn't matter what vg val is passed
        within_grp_case(blocks, vg=None)
                
    # Start recursive loop with vg=0 within group case
    within_grp_case(blocks, vg=0)

    # Return vgs so that that they
    # got neatly from 0 to n-1
    return LabelEncoder().fit_transform(vgs)


from setuptools import setup, find_packages

from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(name='bp-neurotools',
      version='0.22',
      long_description=long_description,
      long_description_content_type='text/markdown',
      description='General helper functions for working with neuroimaging data.',
      url='http://github.com/sahahn/neurotools',
      author='Sage Hahn',
      author_email='sahahn@uvm.edu',
      license='MIT',
      python_requires=">=3.7",
      packages=find_packages(),
      install_requires=[
          'scikit-learn>=1',
          'scipy',
          'numpy',
          'pandas>=1.1',
          'nibabel>=3.2.1',
          'joblib',
          'statsmodels>=0.13.2',
          'matplotlib>=3.3.4',
          'nilearn>=0.8.1',
          'networkx'
      ],
      test_suite='pytest',
      tests_require=['pytest', 'coverage'],
      zip_safe=False)



import numpy as np

def drop_top_x_outliers(data, subjects=None, top=50):
    '''Drop a fixed number of outliers from a set of loaded subjects data,
    based on the absolute mean of each subject's data.

    Parameters
    -----------
    data : numpy array
        Loaded 2D numpy array with shape # of subjects x data points,
        representing the loaded data in which to drop outliers from.

    subjects : numpy array or None, optional
        A corresponding list or array like of subjects with the same
        length as the first dimension of the passed data. If passed,
        then will return as a modified subject array representing
        the new subject list with dropped outliers excluded.
        If not passed, i.e., left as default None,
        will only return the modified data.

        ::

            default = None

    top : int, optional
        The number of subjects to drop.

        ::

            default = 50

    Returns
    --------
    data : numpy array
        The loaded 2D data, but with `top` subjects data removed.

    subjects : numpy array
        If passed originally as None, only data will be returned,
        otherwise this will represent the new list of kept subjects,
        with `top` removed.

    '''

    abs_data = np.abs(data)
    m = np.mean(abs_data, axis=1)
    to_drop = np.argsort(m)[-top:]
    to_keep = np.ones(m.shape, dtype='bool')
    to_keep[to_drop] = 0

    if subjects is not None:
        return data[to_keep], np.array(subjects)[to_keep]
    
    # Otherwise return just data
    return data[to_keep]


from .plot import (plot_surf_hemi, plot_surf_collage, 
                   plot_surf_vol_collage,
                   plot)
from .ref import SurfRef, VolRef

__all__ = ['plot_surf_hemi',
           'plot_surf_collage', 
           'plot_surf_vol_collage',
           'plot', 'SurfRef', 'VolRef']


import warnings

def _get_print(verbose, _print=None):

    # Top behavior should be, if already print
    # init'ed keep that
    if not _print is None:
        return _print

    # Helper for muted print
    if verbose is None:
        return _get_print(verbose=-10)

    def _print(*args, **kwargs):

        if 'level' in kwargs:
            level = kwargs.pop('level')
        else:
            level = 1

        # Skip any print if the verbose is under 0
        if verbose < 0:
            return

        # Use warnings for level = 0
        # If warn, don't also print
        if level == 0:

            # Conv print to str - then warn
            sep = ' '
            if 'sep' in kwargs:
                sep = kwargs.pop('sep')
            as_str = sep.join(str(arg) for arg in args)
            warnings.warn(as_str)

        elif verbose >= level:
            print(*args, **kwargs, flush=True)

    return _print


from sklearn.preprocessing import OrdinalEncoder

import warnings
with warnings.catch_warnings():
    warnings.simplefilter(action='ignore', category=FutureWarning)
    import statsmodels.api as sm

def run_mixed_model(df, fixed_effects_vars, random_effects_vars):
    '''Wrapper code for running a mixed linear model with either one random effect
    or two nested random effects.

    Note: Since the mixed model does not support NaN data, any rows
    with any missing data will be dropped, and the model run on
    just subjects without missing data.

    Parameters
    -----------
    df : pandas DataFrame
        A DataFrame containing columns for the fixed effects
        and the random effects.

        The dataframe should also contain a single column with
        name 'data' which represents the dependent variable,
        e.g., `data ~ fixed_effects_vars`

    fixed_effects_vars : list of str
        List of fixed effect variables, where if any
        variables are categorical they should be wrapped in `C(variable_name)`.

    random_effects_vars : list of str or str
        Either a single random effect or a list with two
        random effect variables specifying that multiple
        layers of nesting should be done. In the case
        where two variables should be nested, the first element
        of the list represents the larger group and the second
        the nested group within the larger group.

        As of right now only these two options are supported.
    
    Returns
    -------
    fitted_model : MixedLMResults
        Returns an instance of MixedLMResults with fitted results.

    '''

    # Check for missing data, if any, drop those subjects
    valid = df[~df['data'].isnull()].index
    df = df.loc[valid]
    
    # If just one random effect specified
    if len(random_effects_vars) == 1:
        if isinstance(random_effects_vars, str):
            random_effects_vars = [random_effects_vars]
        
        re_formula, vc_formula = None, None
    
    # Specify nested structure if two random effects
    elif len(random_effects_vars) == 2:
        
        re_formula = "1"
        vc_formula = {random_effects_vars[1]: "0+C(" + random_effects_vars[1] + ")"}

    else:
         raise RuntimeError('More than 2 random effects are not currently supported.')
    
    # Make sure random effects are ordinalized
    df[random_effects_vars] = OrdinalEncoder().fit_transform(df[random_effects_vars])
    
    # Set base formula as data explained by fixed effect vars
    base = 'data ~ ' + fixed_effects_vars[0]
    base += ' + '.join([''] + fixed_effects_vars[1:])
    
    # Get model from formula
    model =\
        sm.MixedLM.from_formula(base, groups=random_effects_vars[0],
                                re_formula=re_formula,
                                vc_formula=vc_formula, data=df)
    result = model.fit()
    return result

'''
def run_full_model(data, df, random_effects_vars, indep_var, fixed_effects_vars):
    
    results = []
    start_time = time.time()

    for i in range(data.shape[1]):

        print('Vertex:', i, flush=True)
        start_run = time.time()

        df['data'] = data[:, i]  
        result = _run_model(df, random_effects_vars,
                            indep_var, fixed_effects_vars)

        print('single run time:', time.time() - start_run, flush=True)
        results.append(result)        

    print('full run time:', time.time() - start_time, flush=True)
    return results
'''



import  numpy as np

from ..permutations import (_get_contrast, _get_perm_matrix,
                            _proc_input, permuted_v)

def test_get_contrast_1():

    t = np.ones((10, 1))
    c = np.zeros((10, 2))

    contrast = _get_contrast(t, c)

    assert contrast.shape == (3, 1)
    assert contrast[0][0] == 1
    assert contrast[1][0] == 0
    assert contrast[2][0] == 0

def test_get_contrast_2():

    t = np.ones((10, 1))
    c = np.zeros((10, 1))

    contrast = _get_contrast(t, c)

    assert contrast.shape == (2, 1)
    assert contrast[0][0] == 1
    assert contrast[1][0] == 0

def test_get_contrast_3():

    t = np.ones((10, 1))
    c = np.zeros((10, 0))

    contrast = _get_contrast(t, c)

    assert contrast.shape == (1, 1)
    assert contrast[0][0] == 1

def test_get_perm_matrix():

    p_structure = np.array([1, 1, 1, 2, 2, 2])

    p_set = _get_perm_matrix(p_structure, None)
    
    assert 'bool' in p_set.dtype.name
    assert p_set.shape == ((6, 6))

def test_proc_input_1():

    tested_vars = np.ones((10, 1))
    confounding_vars = np.ones((10, 1))

    X, Z, _ = _proc_input(tested_vars, confounding_vars)

    np.testing.assert_almost_equal(np.sum(X), 0)
    np.testing.assert_almost_equal(np.sum(Z), -3.162277660168379)

def test_permuted_v_one_group():

    pvals, original_scores, h0_vmax =\
        permuted_v(tested_vars=np.ones((4)),
                   target_vars=np.random.random((4, 5)),
                   confounding_vars=np.random.random((4, 2)),
                   permutation_structure=np.array([1, 1, 2, 2]),
                   n_perm=10,
                   two_sided_test=True,
                   within_grp=True,
                   random_state=None,
                   use_tf=False,
                   n_jobs=1,
                   verbose=0,
                   dtype=None,
                   use_z=False,
                   demean_confounds=True,
                   min_vg_size=None)

    assert len(pvals) == 5
    assert len(original_scores) ==  5
    assert len(h0_vmax) == 10

def test_permuted_v_rand_vals():

    pvals, original_scores, h0_vmax =\
        permuted_v(tested_vars=np.random.random((4)),
                   target_vars=np.random.random((4, 5)),
                   confounding_vars=np.random.random((4, 2)),
                   permutation_structure=np.array([1, 1, 2, 2]),
                   n_perm=10,
                   two_sided_test=True,
                   within_grp=True,
                   random_state=None,
                   use_tf=False,
                   n_jobs=1,
                   verbose=0,
                   dtype=None,
                   use_z=False,
                   demean_confounds=True,
                   min_vg_size=None)

    assert len(pvals) == 5
    assert len(original_scores) ==  5
    assert len(h0_vmax) == 10

def test_permuted_v_rand_vals_z():

    pvals, original_scores, h0_vmax =\
        permuted_v(tested_vars=np.random.random((4)),
                   target_vars=np.random.random((4, 5)),
                   confounding_vars=np.random.random((4, 2)),
                   permutation_structure=np.array([1, 1, 2, 2]),
                   n_perm=10,
                   two_sided_test=True,
                   within_grp=True,
                   random_state=None,
                   use_tf=False,
                   n_jobs=1,
                   verbose=0,
                   dtype=None,
                   use_z=True,
                   demean_confounds=True,
                   min_vg_size=None)

    assert len(pvals) == 5
    assert len(original_scores) ==  5
    assert len(h0_vmax) == 10

def test_permuted_v_rand_vals_float32():

    pvals, original_scores, h0_vmax =\
        permuted_v(tested_vars=np.random.random((4)),
                   target_vars=np.random.random((4, 5)),
                   confounding_vars=np.random.random((4, 2)),
                   permutation_structure=np.array([1, 1, 2, 2]),
                   n_perm=10,
                   two_sided_test=True,
                   within_grp=True,
                   random_state=None,
                   use_tf=False,
                   n_jobs=1,
                   verbose=0,
                   dtype='float32',
                   use_z=False,
                   demean_confounds=True,
                   min_vg_size=None)

    assert len(pvals) == 5
    assert len(original_scores) ==  5
    assert 'float32' in original_scores.dtype.name
    assert len(h0_vmax) == 10

def test_permuted_v_rand_vals_float32_demean_dif_no_perm():

    tested_vars=np.random.random((4))
    target_vars=np.random.random((4, 5))
    confounding_vars=np.random.random((4, 2))
    permutation_structure=np.array([1, 1, 2, 2])


    _, original_scores1, _ =\
        permuted_v(tested_vars=tested_vars,
                   target_vars=target_vars,
                   confounding_vars=confounding_vars,
                   permutation_structure=permutation_structure,
                   n_perm=0,
                   two_sided_test=True,
                   within_grp=True,
                   random_state=None,
                   use_tf=False,
                   n_jobs=1,
                   verbose=0,
                   dtype='float32',
                   use_z=False,
                   demean_confounds=False,
                   min_vg_size=None)

    _, original_scores2, _ =\
        permuted_v(tested_vars=tested_vars,
                   target_vars=target_vars,
                   confounding_vars=confounding_vars,
                   permutation_structure=permutation_structure,
                   n_perm=0,
                   two_sided_test=True,
                   within_grp=True,
                   random_state=None,
                   use_tf=False,
                   n_jobs=1,
                   verbose=0,
                   dtype='float32',
                   use_z=False,
                   demean_confounds=True,
                   min_vg_size=None)

    assert not np.array_equal(original_scores1, original_scores2)

def test_permuted_v_rand_vals_demean_dif_no_perm():

    tested_vars=np.random.random((4))
    target_vars=np.random.random((4, 5))
    confounding_vars=np.random.random((4, 2))
    permutation_structure=np.array([1, 1, 2, 2])


    _, original_scores1, _ =\
        permuted_v(tested_vars=tested_vars,
                   target_vars=target_vars,
                   confounding_vars=confounding_vars,
                   permutation_structure=permutation_structure,
                   n_perm=0,
                   two_sided_test=True,
                   within_grp=True,
                   random_state=None,
                   use_tf=False,
                   n_jobs=1,
                   verbose=0,
                   dtype=None,
                   use_z=False,
                   demean_confounds=False,
                   min_vg_size=None)

    _, original_scores2, _ =\
        permuted_v(tested_vars=tested_vars,
                   target_vars=target_vars,
                   confounding_vars=confounding_vars,
                   permutation_structure=permutation_structure,
                   n_perm=0,
                   two_sided_test=True,
                   within_grp=True,
                   random_state=None,
                   use_tf=False,
                   n_jobs=1,
                   verbose=0,
                   dtype=None,
                   use_z=False,
                   demean_confounds=True,
                   min_vg_size=None)

    assert not np.array_equal(original_scores1, original_scores2)

def test_permuted_v_rand_vals_dtype_compare():

    tested_vars=np.random.random((4))
    target_vars=np.random.random((4, 5))
    confounding_vars=np.random.random((4, 2))
    permutation_structure=np.array([1, 1, 2, 2])

    _, original_scores1, _ =\
        permuted_v(tested_vars=tested_vars,
                target_vars=target_vars,
                confounding_vars=confounding_vars,
                permutation_structure=permutation_structure,
                n_perm=0,
                two_sided_test=True,
                within_grp=True,
                random_state=None,
                use_tf=False,
                n_jobs=1,
                verbose=0,
                dtype='float64',
                use_z=False,
                demean_confounds=True,
                min_vg_size=None)

    _, original_scores2, _ =\
        permuted_v(tested_vars=tested_vars,
                target_vars=target_vars,
                confounding_vars=confounding_vars,
                permutation_structure=permutation_structure,
                n_perm=0,
                two_sided_test=True,
                within_grp=True,
                random_state=None,
                use_tf=False,
                n_jobs=1,
                verbose=0,
                dtype='float32',
                use_z=False,
                demean_confounds=True,
                min_vg_size=None)

    # All close, but not equal
    assert np.allclose(original_scores1, original_scores2,
                        rtol=1e-02, atol=1e-05)
    assert not np.array_equal(original_scores1, original_scores2)


def test_permuted_v_rand_vals_min_vg_size():

    tested_vars=np.random.random((6))
    target_vars=np.random.random((6, 5))
    confounding_vars=np.random.random((6, 2))
    permutation_structure=np.array([1, 1, 1, 3, 3, 2])

    _, original_scores1, _ =\
        permuted_v(tested_vars=tested_vars,
                   target_vars=target_vars,
                   confounding_vars=confounding_vars,
                   permutation_structure=permutation_structure,
                   n_perm=0,
                   two_sided_test=True,
                   within_grp=True,
                   random_state=None,
                   use_tf=False,
                   use_z=False,
                   demean_confounds=True,
                   min_vg_size=None)

    _, original_scores2, _ =\
        permuted_v(tested_vars=tested_vars,
                   target_vars=target_vars,
                   confounding_vars=confounding_vars,
                   permutation_structure=permutation_structure,
                   n_perm=0,
                   two_sided_test=True,
                   within_grp=True,
                   random_state=None,
                   use_z=False,
                   demean_confounds=True,
                   min_vg_size=2)

    # All close, but not equal
    assert not np.array_equal(original_scores1, original_scores2)

def test_permuted_v_rand_vals_min_vg_size2():

    tested_vars=np.random.random((6))
    target_vars=np.random.random((6, 5))
    confounding_vars=np.random.random((6, 2))
    permutation_structure=np.array([1, 1, 1, 3, 3, 2])

    _, original_scores1, _ =\
        permuted_v(tested_vars=tested_vars,
                   target_vars=target_vars,
                   confounding_vars=confounding_vars,
                   permutation_structure=permutation_structure,
                   n_perm=0,
                   two_sided_test=True,
                   within_grp=True,
                   random_state=None,
                   use_tf=False,
                   use_z=False,
                   demean_confounds=True,
                   min_vg_size=2)

    # Should get same result from manual drop
    tested_vars = tested_vars[:-1]
    target_vars = target_vars[:-1]
    confounding_vars = confounding_vars[:-1]
    permutation_structure  = permutation_structure[:-1]

    _, original_scores2, _ =\
    permuted_v(tested_vars=tested_vars,
                target_vars=target_vars,
                confounding_vars=confounding_vars,
                permutation_structure=permutation_structure,
                n_perm=0,
                two_sided_test=True,
                within_grp=True,
                random_state=None,
                use_z=False,
                demean_confounds=True,
                min_vg_size=None)

    # All close, but not equal
    assert np.array_equal(original_scores1, original_scores2)

def test_permuted_v_rand_vals_no_covars():

    pvals, original_scores, h0_vmax =\
        permuted_v(tested_vars=np.random.random((4)),
                   target_vars=np.random.random((4, 5)),
                   confounding_vars=None,
                   permutation_structure=np.array([1, 1, 2, 2]),
                   n_perm=10,
                   two_sided_test=True,
                   within_grp=True,
                   random_state=None,
                   use_tf=False,
                   n_jobs=1,
                   verbose=0,
                   dtype=None,
                   use_z=False,
                   demean_confounds=True,
                   min_vg_size=None)

    assert len(pvals) == 5
    assert len(original_scores) ==  5
    assert len(h0_vmax) == 10

def test_permuted_v_rand_vals_no_covars_model_intercept():

    pvals, original_scores, h0_vmax =\
        permuted_v(tested_vars=np.random.random((4)),
                   target_vars=np.random.random((4, 5)),
                   confounding_vars=None,
                   permutation_structure=np.array([1, 1, 2, 2]),
                   n_perm=10,
                   two_sided_test=True,
                   within_grp=True,
                   random_state=None,
                   use_tf=False,
                   n_jobs=1,
                   verbose=0,
                   dtype=None,
                   use_z=False,
                   model_intercept=True,
                   demean_confounds=True,
                   min_vg_size=None)

    assert len(pvals) == 5
    assert len(original_scores) ==  5
    assert len(h0_vmax) == 10


"""
Dictionary learning estimator: Perform a map learning algorithm by learning
a temporal dense dictionary along with sparse spatial loadings, that
constitutes output maps
"""

# Author: Arthur Mensch
# License: BSD 3 clause

import warnings

import numpy as np
from sklearn.decomposition import dict_learning_online
from sklearn.linear_model import Ridge

from .base import BaseDecomposition
from .canica import CanICA

# check_input=False is an optimization available in sklearn.
sparse_encode_args = {'check_input': False}

def _compute_loadings(components, data):
    ridge = Ridge(fit_intercept=None, alpha=1e-8)
    ridge.fit(components.T, np.asarray(data.T))
    loadings = ridge.coef_.T

    S = np.sqrt(np.sum(loadings ** 2, axis=0))
    S[S == 0] = 1
    loadings /= S[np.newaxis, :]
    return loadings


class DictLearning(BaseDecomposition):

    def __init__(self,
                 n_components=20,
                 n_epochs=1,
                 alpha=10,
                 reduction_ratio='auto',
                 dict_init=None,
                 random_state=None,
                 batch_size=20,
                 method="cd",
                 n_jobs=1,
                 verbose=0):

        BaseDecomposition.__init__(self, n_components=n_components,
                                   random_state=random_state,
                                   n_jobs=n_jobs,
                                   verbose=verbose)
        self.n_epochs = n_epochs
        self.batch_size = batch_size
        self.method = method
        self.alpha = alpha
        self.reduction_ratio = reduction_ratio
        self.dict_init = dict_init

    def _init_dict(self, data):

        if self.dict_init is not None:
            components = self.dict_init

        else:
            canica = CanICA(n_components=self.n_components,
                            do_cca=True,
                            threshold=float(self.n_components),
                            n_init=1,
                            random_state=self.random_state,
                            n_jobs=self.n_jobs, verbose=self.verbose)
            
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", UserWarning)
                canica._raw_fit(data)

            components = canica.components_

        S = (components ** 2).sum(axis=1)
        S[S == 0] = 1
        components /= S[:, np.newaxis]
        self.components_init_ = components

    def _init_loadings(self, data):
        self.loadings_init_ = _compute_loadings(self.components_init_, data)

    def _raw_fit(self, data):
        """Helper function that directly process unmasked data
        Parameters
        ----------
        data : ndarray,
            Shape (n_samples, n_features)
        """
        if self.verbose:
            print('[DictLearning] Learning initial components')
        self._init_dict(data)

        _, n_features = data.shape

        if self.verbose:
            print('[DictLearning] Computing initial loadings')
        self._init_loadings(data)

        dict_init = self.loadings_init_

        n_iter = ((n_features - 1) // self.batch_size + 1) * self.n_epochs

        if self.verbose:
            print('[DictLearning] Learning dictionary')

        self.components_, _ = dict_learning_online(
            data.T, self.n_components, alpha=self.alpha, n_iter=n_iter,
            batch_size=self.batch_size, method=self.method,
            dict_init=dict_init, verbose=max(0, self.verbose - 1),
            random_state=self.random_state, return_code=True, shuffle=True,
            n_jobs=1)
        self.components_ = self.components_.T

        # Unit-variance scaling
        S = np.sqrt(np.sum(self.components_ ** 2, axis=1))
        S[S == 0] = 1
        self.components_ /= S[:, np.newaxis]

        # Flip signs in each composant so that positive part is l1 larger
        # than negative part. Empirically this yield more positive looking maps
        # than with setting the max to be positive.
        for component in self.components_:
            if np.sum(component > 0) < np.sum(component < 0):
                component *= -1

        return self